/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                kruskal.h                         *
 *                                                  *
 *  April 13, 2003                                  *
 ****************************************************/

#ifndef KRUSKAL_H
#define KRUSKAL_H

#include <vector>

#include "../data_structures/edgeVector.h"
#include "../data_structures/mst.h"

/**
   Represents Kruskal's algorithm for determining a Minimum Spanning Tree
   of a weighted graph and the required union/find data structure.
*/
template < typename EdgeType = Edge >
class Kruskal
{
 public:
    /**
       Computes a MST of a graph.
       @param graph a reference to an EdgeVector which represents the graph
       @param result a reference to a MST object which stores the result
    */
    Kruskal(EdgeVector<EdgeType> &graph, MST &result)
	: _graph( graph ), _result( result )
	{
	    computeMST();
	}

    /**
       The destructor.
       The given graph is deleted.
    */
    ~Kruskal() {delete &_graph;}
    
 private:
    /** Computes a MST of the graph. */
    void computeMST();
    
    /** Initializes the union/find data structure. */
    void initUnionFind();


    /**
       Performs a find operation.
       Path compression is applied.
       @param node the identifier of the node whose set should be determined
       @return the identifier of the canonical node which represents the set
               which "node" belongs to
    */
    NodeID find(NodeID node);

    /**
       Performs a union operation.
       @return true iff node1 and node2 have belonged to different sets
    */
    bool unite(NodeID node1, NodeID node2);


    /** Reference to the given graph. */
    EdgeVector<EdgeType> &_graph;

    /** Reference to a MST object which stores the result. */
    MST &_result;

    /**
       A vector which contains for each node the identifier of the parent node.
       The canonical node of a set (= the root of a tree) points to itself.
    */
    std::vector<NodeID> _parent;

    /**
       A vector which contains for each node the height of the belonging tree.
       Only the values of canonical nodes are relevant.
    */
    std::vector<char> _height;

    /** Counts the number of edges which have been added to the resulting MST */
    EdgeCount edgesAddedToResult;
    
};


/**
   Performs a find operation.
   Path compression is applied.
   @param node the identifier of the node whose set should be determined
   @return the identifier of the canonical node which represents the set
           which "node" belongs to
*/
template < typename EdgeType >
inline NodeID Kruskal<EdgeType>::find(NodeID node)
{
    // find the root
    NodeID root = node;
    while (_parent[root] != root) root = _parent[root];

    // apply path compression
    NodeID tmp;
    while (_parent[node] != node) {
	tmp = _parent[node];
	_parent[node] = root;
	node = tmp;
    }
    
    return root;
}

/**
   Performs a union operation.
   @return true iff node1 and node2 have belonged to different sets
*/
template < typename EdgeType >
inline bool Kruskal<EdgeType>::unite(NodeID node1, NodeID node2)
{
    NodeID root1 = find(node1);
    NodeID root2 = find(node2);

    // A cycle would be produced. Therefore the union operation is cancelled.
    if (root1 == root2) return false;

    // Add the smaller tree to the bigger tree
    if (_height[root1] < _height[root2]) {
	_parent[root1] = root2;
    }
    else {
	_parent[root2] = root1;
	
	// Increment the height of the resulting tree if both trees have
	// the same height
	if (_height[root1] == _height[root2]) _height[root1]++;
    }
    return true;
}


#endif // KRUSKAL_H
