/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                buckets.h                         *
 *                                                  *
 *  April 23, 2003                                  *
 ****************************************************/

#ifndef BUCKETS_H
#define BUCKETS_H

#include <vector>

#include "../stxxl/containers/stack.h"
#include "../data_structures/sparingStack.h"

#include "../data_structures/edgeVector.h"
#include "../data_structures/mst.h"


/**
   Represents an algorithm for reducing the number of nodes
   of a graph (in order to compute a minimum spanning tree)
   and the required data structures.
   This implementation uses several buckets.

   template parameter:
   The number of nodes per bucket.
   This value doesn't apply to the first bucket, because
   it contains the edges of the first "_noOfNodesInIntMem"
   nodes.
*/
template <NodeCount nodesPerBucket = 1>
class Buckets
{
public:
    typedef int BucketID;
    
private:
    /** The type of an external bucket which contains the edges of several nodes. */
    typedef stxxl::STACK_GENERATOR<RelabeledEdge,
				   stxxl::external,
				   stxxl::grow_shrink2,
				   DS_EXT_PAGE_SIZE,
				   DS_EXT_BLOCK_SIZE>::result EdgesOfSeveralNodes;

    /** The type of an internal bucket which contains the edges of one node. */
    typedef REWS_SparingStack<DS_INT_EDGES_PER_BLOCK, DS_INT_NO_OF_BLOCKS> EdgesOfOneNode;
    
    /** The type of the common pool of the internal buckets. */
    typedef CommonPoolOfBlocks<RelabeledEdgeWithoutSource,
			       DS_INT_EDGES_PER_BLOCK,
			       DS_INT_NO_OF_BLOCKS> PoolEdgesOfOneNode;
    
public:
    /** Computes the number of external buckets that are needed. */
    static BucketID noOfExtBuckets(NodeCount noOfNodes, NodeCount noOfNodesInIntMem);
    
	
    /**
       Reduces the number of nodes of a graph (in order to
       compute a minimum spanning tree).
       @param graph a reference to an EdgeVector which represents the graph
       @param result a reference to a MST object which stores the result
       @param noOfBuckets the number of external buckets which should be used
       @param noOfNodesInIntMem the number of nodes which fit in internal memory
    */
    Buckets(EdgeVector<Edge> &graph, MST &result,
	    BucketID noOfBuckets, NodeCount noOfNodesInIntMem)
	: _graph( graph ), _result( result ), _extBuckets( noOfBuckets-1 ),
	  _noOfNodesInIntMem( noOfNodesInIntMem ),
	  _prefetchPool( DS_EXT_PREFETCH_POOL ), _writePool( DS_EXT_WRITE_POOL )
	{
	    initBuckets();
	    reduceNodes();
	}
    
    /**
       Returns a pointer to the first external bucket which contains the
       edges of the nodes which fit in internal memory.
    */
    InternalMemoryBucket* getIntMemBucket() {return _firstExtBucket;}

    /**
       Adds the given edge to the appropriate external bucket or
       to the appropriate internal bucket if the edge belongs
       to the external bucket which is processed at the moment.
    */
    void add(const RelabeledEdge &edge);
    
private:
    /** Reference to the given graph. */
    EdgeVector<Edge> &_graph;

    /** Reference to a MST object which stores the result. */
    MST &_result;

    /** The number of nodes which fit in internal memory. */
    NodeCount _noOfNodesInIntMem;

    /**
       The first external bucket which contains the
       edges of the nodes which fit in internal memory.
    */
    InternalMemoryBucket *_firstExtBucket;
    
    /**
       The external buckets.
       Each bucket contains the edges of several nodes.
    */
    std::vector<EdgesOfSeveralNodes*> _extBuckets;

    /** The prefetch pool that is used by the external buckets. */
    stxxl::prefetch_pool<EdgesOfSeveralNodes::block_type> _prefetchPool;

    /** The write pool that is used by the external buckets. */
    stxxl::write_pool<EdgesOfSeveralNodes::block_type> _writePool;

    /**
       The internal buckets.
       Each bucket contains the edges of one node.
    */
    EdgesOfOneNode _intBuckets[nodesPerBucket];
    
    /**
       The identifier of the first node in the current external bucket will be used
       to compute the internal bucket index of a node.
    */
    NodeID _firstNodeIDofCurrentBucket;

    /**
       Initializes the external buckets and distributes the edges to the
       external buckets.
    */
    void initBuckets();

    /**
       Reduces the number of nodes.
       Only the first bucket "survives".
    */
    void reduceNodes();

    /**
       Returns the ID of the bucket which contains the edges
       of the given node.
       The first external bucket has the ID -1 because it is a
       special case. The second bucket is the first element
       of _extBuckets and has the ID 0 and so on.
    */
    BucketID bucketID(NodeID nodeID) const;

    /** Adds the given edge to the appropriate external bucket. */
    void addToExternalBucket(const RelabeledEdge &edge);

    /** Adds the given edge to the bucket which is specified by newBucketID. */
    void addToExternalBucket(const RelabeledEdge &edge, BucketID newBucketID);
};


/**
   Returns the ID of the bucket which contains the edges
   of the given node.
   The first external bucket has the ID -1 because it is a
   special case. The second bucket is the first element
   of _extBuckets and has the ID 0 and so on.
*/
template <NodeCount nodesPerBucket>
inline typename Buckets<nodesPerBucket>::BucketID Buckets<nodesPerBucket>::bucketID(NodeID nodeID)
const
{
    if (nodeID < _noOfNodesInIntMem) return -1;
    return (nodeID - _noOfNodesInIntMem) / nodesPerBucket;
}

/** Adds the given edge to the appropriate external bucket. */
template <NodeCount nodesPerBucket>
inline void Buckets<nodesPerBucket>::addToExternalBucket(const RelabeledEdge &edge)
{
    if ( edge.isSelfLoop() ) return; // Self loops can be ignored.
    addToExternalBucket( edge, bucketID(edge.source()) );
}

/** Adds the given edge to the bucket which is specified by newBucketID. */
template <NodeCount nodesPerBucket>
inline void Buckets<nodesPerBucket>::addToExternalBucket(const RelabeledEdge &edge,
							 BucketID newBucketID)
{
    if (newBucketID == -1) {
	// special case
	_firstExtBucket->push_back( edge );
    }
    else {
	_extBuckets[newBucketID]->push( edge );
    }
}

/**
   Adds the given edge to the appropriate external bucket or
   to the appropriate internal bucket if the edge belongs
   to the external bucket which is processed at the moment.
*/
template <NodeCount nodesPerBucket>
inline void Buckets<nodesPerBucket>::add(const RelabeledEdge &edge)
{
    if (edge.source() >= _firstNodeIDofCurrentBucket) {
	// The relabeled edge still belongs to the current external bucket.
	// Therefore it is added to the appropriate internal bucket of one
	// single node.
	_intBuckets[edge.source() - _firstNodeIDofCurrentBucket]
	    .push(edge);
    }
    else {
	// The relabeled edge now belongs to another external bucket.
	// Therefore it is added to the appropriate external bucket.
	addToExternalBucket( edge, bucketID(edge.source()) );
    }
}

#endif // BUCKETS_H
