/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                duplicatesRemover.h               *
 *                                                  *
 *  May 20, 2003                                    *
 ****************************************************/

#ifndef DUPLICATESREMOVER_H
#define DUPLICATESREMOVER_H

/**
   A container which can be used as an 'intermediate station' in order
   to remove multiple edges. If there is more than one edge with the
   same target node (the source node is always the same during one pass),
   only the edge with minimum weight is preserved.
   If the capacity of the container is exhausted, further edges which can't
   be stored are output directly.
*/
template <typename SuperContainer>
class DuplicatesRemover
{
 public:
    /**
       Creates a DuplicatesRemover.
       @param superContainer a pointer to the super container, i.e. the container
                             that uses this DuplicatesRemover (either a Buckets or
			     a PQueue object). This is required in order to be able
			     to put the edges to the super container when the capacity
			     of this container is exhausted or when this container is
			     cleared.
    */
    DuplicatesRemover(SuperContainer *superContainer)
	: _superContainer( superContainer ), _size( 0 )
	{}

    /**
       Adds an edge to the container.
       If an edge with the same target node is already in the container,
       the new edge replaces the old one if it has a lower weight, otherwise
       it is discarded.
    */
    void insert(const RelabeledEdgeWithoutSource &edge, NodeID source) {
	int index = find( edge.target() );
	if ( empty(index) ) {
	    // There is NO edge with the same target node.
	    if (_size == _maxSize) {
		// The container is full, so the current edge has to be
		// output directly.
		_superContainer->add( RelabeledEdge(edge, source) );
	    }
	    else {
		// Add the edge to the container
		_hashMap[index] = HashMapElement( edge, _size );
		_edges[_size++] = edge;
	    }
	}
	else {
	    // There is another edge with the same target node.
	    // So this a duplicate !
	    DS_LOGGING2( logging.duplicates().increase() );
	    if (edge < _hashMap[index].first) {
		// The current edge has a lower weight than the existing edge.
		// Therefore the existing edge is replaced.
		_edges[ _hashMap[index].second ] = edge;
		_hashMap[index].first = edge;
	    }
	}
    }

    /** Clears the container. The edges are written to the appropriate buckets. */
    void clear(NodeID source) {
	while ( _size > 0 ) {
	    clearLocal( _edges[--_size].target() );
	    _superContainer->add( RelabeledEdge(_edges[_size], source) );
	}
    }
    
 private:
    /** The capacity of the container. */
    static const EdgeCount _maxSize = 1024;

    /** The size of the hash map. */
    static const EdgeCount _hashMapSize = 2 * _maxSize;
    
    /**
       The type of a hashMap-element.
       The first component stores the edge, the second the index in the
       array '_edges'.
    */
    typedef std::pair<RelabeledEdgeWithoutSource, int> HashMapElement;

    /**
       This array contains all edges without vacancies.
       This is useful in order to clear the container without traversing
       the whole hashMap.
    */
    RelabeledEdgeWithoutSource _edges[_maxSize];

    /** The hashMap which contains the edges. */
    HashMapElement _hashMap[_hashMapSize];

    /** The number of edges which are stored in this container. */
    EdgeCount _size;

    /**
       A pointer to the super container, i.e. the container that uses
       this DuplicatesRemover (either a Buckets or a PQueue object).
       This is required in order to be able to put the edges to the
       super container when the capacity of this container is exhausted
       or when this container is cleared.
    */
    SuperContainer *_superContainer;

    /** Returns the hash value for the given node ID. */
    int hashFunction(NodeID x) const {
	return x % _hashMapSize;	
    }

    /**
       Returns the position of an edge with the given node ID in the hashMap.
       If no appropriate edge is found, the position where such an edge should
       be stored is returned.
    */
    int find(NodeID x) const {
	int i = hashFunction(x);
	while ((_hashMap[i].first.target() != x) && ( ! empty(i) ))
	    i = (i+1) % _hashMapSize;
	return i;
    }

    /**
       Removes entries from the hashMap beginning with the hash value of
       the given node ID and ending with the first empty entry which is found.
    */
    void clearLocal(NodeID x) {
	int i = hashFunction(x);
	while ( ! empty(i) ) {
	    _hashMap[i].first = RelabeledEdgeWithoutSource();
	    i = (i+1) % _hashMapSize;
	}
    }

    /** Returns true iff the given position in the hashMap is empty. */
    bool empty(int index) const {
	if ( (_hashMap[index].first.originalSource() == 0) &&
	     (_hashMap[index].first.originalTarget() == 0) ) return true;
	return false;
    }

};

#endif // DUPLICATESREMOVER_H
