/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                edge.h                            *
 *                                                  *
 *  April 13, 2003                                  *
 ****************************************************/


#ifndef EDGE_H
#define EDGE_H

#include <iostream>

typedef unsigned int NodeID;
typedef unsigned int EdgeWeight;
typedef unsigned long long int EdgeWeightBig; // used for the sum of edge weights

/**
   Represents an edge without the source vertex.
   This makes sense when the source vertex is clear from the context.
*/
class EdgeWithoutSource
{
    /** Compares two edges by weight. */
    friend bool operator<(const EdgeWithoutSource &e1,
			  const EdgeWithoutSource &e2) {
	return e1._weight<e2._weight;
    }

    /**
       Returns true iff two EdgesWithoutSource are identical,
       i.e. the targets and the weights are equal.
    */
    friend bool operator==(const EdgeWithoutSource &e1,
			   const EdgeWithoutSource &e2) {
	return (e1._target == e2._target) && (e1._weight == e2._weight);
    }
    
    /** Writes a string representation of an EdgeWithoutSource to an output stream. */
    friend std::ostream& operator<<( std::ostream& os, const EdgeWithoutSource &e ) {
	os << e.target() << " " << e.weight();
	return os;
    }
    
 public:
    /** Returns an EdgeWithoutSource object with minimum weight. */
    static EdgeWithoutSource minWeight() {return EdgeWithoutSource(0,0);}

    /** Returns an EdgeWithoutSource object with maximum weight. */
    static EdgeWithoutSource maxWeight() {return EdgeWithoutSource(0,0xffffffff);}
    
    /** The default constructor. */
    EdgeWithoutSource( NodeID target = 0, EdgeWeight weight = 0 )
	: _target( target ), _weight( weight )
	{}
    
    /** Returns the identifier of the target vertex. */
    NodeID target() const {return _target;}
    
    /** Returns the weight. */
    EdgeWeight weight() const {return _weight;}
      
 protected:
    /** Sets the target vertex. */
    void setTarget(NodeID target) {_target = target;}
    
 private:
    NodeID _target;
    EdgeWeight _weight; 
};

/**
   Represents a directed and weighted edge.
*/
class Edge : public EdgeWithoutSource
{
    /** Determines if two edges have the same source and the same target vertex. */
    friend bool operator==(const Edge &e1, const Edge &e2) {
	return ( e1.source()==e2.source() ) && ( e1.target()==e2.target() );
    }

    /** Writes a string representation of an Edge to an output stream. */
    friend std::ostream& operator<<( std::ostream& os, const Edge &e ) {
	os << e.source() << " " << e.target() << " " << e.weight();
	return os;
    }
    
 public:
    /** Used by stxxl::ksort in order to sort by weight. */
    typedef EdgeWeight key_type;

    
    /** Returns an Edge object with minimum weight. */
    static Edge minWeight() {return Edge(0,0,0);}

    /** Returns an Edge object with maximum weight. */
    static Edge maxWeight() {return Edge(0,0,0xffffffff);}
    

    /** The default constructor. */
    Edge(NodeID source = 0, NodeID target = 0, EdgeWeight weight = 0)
	: _source( source ), EdgeWithoutSource( target, weight )
	{}

    /** Returns the identifier of the source vertex. */
    NodeID source() const {return _source;}

    /** Swaps the source and the target vertex. */
    void swap() {NodeID tmp = _source; _source = target(); setTarget(tmp);}

    /**
       Returns true iff this Edge is a self loop, i.e. the source and the
       target vertices are identical.
    */
    bool isSelfLoop() const {return (source() == target());}


 private:
    NodeID _source;
};



/**
   Extracts the weight of an edge.
   Used by stxxl::ksort in order to sort by weight.
*/
template <typename EdgeType = Edge>
class GetWeight
{
public:
    typedef Edge::key_type key_type;
    key_type operator() (const EdgeType & obj)
	{
	    return obj.weight();
	}
    static EdgeType min_value(){ return EdgeType::minWeight(); }
    static EdgeType max_value(){ return EdgeType::maxWeight(); }
};



/**
   A StrictWeakOrdering predicate which can be used to compare two edges by
   weight.
   Used by stxxl::sort in order to sort by weight.
*/
template <typename EdgeType = Edge>
class WeightOrdering
{
 public:
    bool operator() (const EdgeType &e1, const EdgeType &e2) const
	{
	    if (e1.weight() < e2.weight()) return true;
	    return false;
	}
    static EdgeType min_value(){ return EdgeType::minWeight(); }
    static EdgeType max_value(){ return EdgeType::maxWeight(); }
};


/**
   A StrictWeakOrdering predicate which can be used to compare two edges first by
   the source vertex, then - if the source vertices are equal - by the target
   vertex and finally by weight.
*/
class SourceTargetWeightOrdering
{
 public:
    bool operator() (const Edge &e1, const Edge &e2) const
	{
	    // first, compare by source
	    if (e1.source() < e2.source()) return true;
	    if (e1.source() > e2.source()) return false;

	    // then, by target
	    if (e1.target() < e2.target()) return true;
	    if (e1.target() > e2.target()) return false;

	    // finally, by weight
	    if (e1.weight() < e2.weight()) return true;
	    return false;
	}

    static Edge min_value() {return Edge(0,0,0);}
    static Edge max_value() {return Edge(0xffffffff,0xffffffff,0xffffffff);}
};


/**
   A StrictWeakOrdering predicate which can be used to compare two edges first by
   the source vertex, then - if the source vertices are equal - by weight.
   Used as comparison type by stxxl::priority_queue. The largest element, which is
   returned by top(), is the edge with the highest source vertex ID and (if there
   are several edges with the same source vertex) minimum weight.
*/
template <typename EdgeType = Edge>
class SourceWeightOrdering
{
 public:
    bool operator() (const EdgeType &e1, const EdgeType &e2) const
	{
	    // first, compare by source
	    if (e1.source() < e2.source()) return true;
	    if (e1.source() > e2.source()) return false;

	    // then, by weight
	    if (e1.weight() > e2.weight()) return true;
	    return false;
	}
    
    static EdgeType min_value(){ return EdgeType::minWeight(); }
    static EdgeType max_value(){ return EdgeType::maxWeight(); }
};

#endif // EDGE_H
