/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                edgeVector.h                      *
 *                                                  *
 *  April 13, 2003                                  *
 ****************************************************/


#ifndef EDGEVECTOR_H
#define EDGEVECTOR_H

#include <iostream>

#include "relabeledEdge.h"

#include "../stxxl/containers/vector.h"

typedef unsigned int EdgeCount;
typedef unsigned int NodeCount;


/**
   Represents a weighted graph by a list of Edges.
*/
template <typename EdgeType = Edge,
	  unsigned int blockSize = DS_DEFAULT_BLOCK_SIZE,
	  unsigned int noOfPages = DS_DEFAULT_NO_OF_PAGES,
	  unsigned int pageSize = DS_DEFAULT_PAGE_SIZE>
class EdgeVector : public stxxl::VECTOR_GENERATOR<EdgeType, pageSize, noOfPages,
						  blockSize>::result
{
    /** Writes a string representation of the edge list to an output stream. */
    friend std::ostream& operator<<( std::ostream& os,
				     EdgeVector<EdgeType,blockSize,noOfPages,pageSize> &el ) {
	os << el.noOfNodes() << " " << el.noOfEdges() << std::endl;
	for (EdgeCount i=0; i<el.noOfEdges(); i++)
	    os << el[i] << std::endl;
	return os;
    }
    
 public:
    /**
       Constructs an EdgeVector.
       The number of nodes is stored and space for the edges is reserved,
       but the edges aren't created. In order to add the edges of the graph,
       use methods like "push_back".
       @param noOfNodes the number of nodes in the graph
       @param noOfEdges the number of edges which space is reserved for
    */
    EdgeVector(NodeCount noOfNodes, EdgeCount noOfEdges)
	: _noOfNodes( noOfNodes )
	{ this->reserve(noOfEdges); }

    /** Returns the number of edges of the graph. */
    EdgeCount noOfEdges() const {return this->size();}

    /** Returns the number of nodes of the graph. */
    NodeCount noOfNodes() const {return _noOfNodes;}

    /** Returns true iff this list is empty, i.e. it contains no edges. */
    bool empty() const {return (noOfEdges() == 0);}

    
    /** Sorts the edges of the graph by weight. */
    void sortByWeight();
    
 private:
    NodeCount _noOfNodes;

};


/**
   The type of the first external bucket which contains the
   edges of the nodes which fit in internal memory.
*/
typedef EdgeVector<RelabeledEdge> InternalMemoryBucket;


#endif // EDGEVECTOR_H
