/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                relabeledEdge.h                   *
 *                                                  *
 *  April 22, 2003                                  *
 ****************************************************/

#ifndef RELABELEDEDGE_H
#define RELABELEDEDGE_H


#include "edge.h"


class RelabeledEdge;
    
/**
   Represents a relabeled edge without the source vertex.
   This makes sense when the source vertex is clear from the context.
*/
class RelabeledEdgeWithoutSource : public EdgeWithoutSource
{
    /**
       Writes a string representation of a RelabeledEdgeWithoutSource
       to an output stream.
    */
    friend std::ostream& operator<<( std::ostream& os,
				     const RelabeledEdgeWithoutSource &e ) {
	os << e.target() << " " << e.weight() << " "
	   << e.originalSource() << " " << e.originalTarget();
	return os;
    }
    
 public:
    /** The default constructor. */
    RelabeledEdgeWithoutSource()
	: EdgeWithoutSource(), _originalSource( 0 ), _originalTarget( 0 )
	{}
    
    /**
       Creates a RelabeledEdgeWithoutSource from a RelabeledEdge.
       The source vertex is thrown away.
    */
    RelabeledEdgeWithoutSource( const RelabeledEdge &edge );

    /** Returns the identifier of the original source vertex. */
    NodeID originalSource() const {return _originalSource;}

    /** Returns the identifier of the original target vertex. */
    NodeID originalTarget() const {return _originalTarget;}
	
 private:
    NodeID _originalSource;
    NodeID _originalTarget;
};


/**
   Represents a relabeled edge. I.e. an edge whose source and target vertices
   have been relabeled and the original indices are stored additionally.
   This class doesn't extend RelabeledEdgeWithoutSource in order to avoid
   multiple inheritance.
*/
class RelabeledEdge : public Edge
{
    /** Writes a string representation of a RelabeledEdge to an output stream. */
    friend std::ostream& operator<<( std::ostream& os, const RelabeledEdge &e ) {
	os << e.source() << " " << e.target() << " " << e.weight() << " "
	   << e.originalSource() << " " << e.originalTarget();
	return os;
    }
    
 public:
    /** Returns a RelabeledEdge object with minimum weight. */
    static RelabeledEdge minWeight() {return RelabeledEdge(0,0,0,0,0);}

    /** Returns a RelabeledEdge object with maximum weight. */
    static RelabeledEdge maxWeight() {return RelabeledEdge(0,0,0xffffffff,0,0);}

    
    /** The default constructor. */
    RelabeledEdge(NodeID source = 0, NodeID target = 0, EdgeWeight weight = 0,
		  NodeID origSource = 0, NodeID origTarget = 0)
	: Edge( source,target,weight ),
	  _originalSource( origSource ), _originalTarget( origTarget )
	{}

    /**
       Creates a relabeled edge from a normal edge.
       The original source resp. target equals the current source resp. target.
    */
    RelabeledEdge(const Edge &edge)
	: Edge( edge ), _originalSource( edge.source() ),
	                _originalTarget( edge.target() )
	{}

    /**
       Creates a relabeled edge from a RelabeledEdgeWithoutSource.
       The source vertex must be provided as second parameter.
       If necessary, source and target are swapped, so that
       source is greater than (or equal to) target.
    */
    RelabeledEdge(const RelabeledEdgeWithoutSource &edge, NodeID newSource)
	: Edge( newSource, edge.target(), edge.weight() ),
	  _originalSource( edge.originalSource() ), _originalTarget( edge.originalTarget() )
	{
	    if (source() < target()) swap(); 
	}
    
    /** Returns the identifier of the original source vertex. */
    NodeID originalSource() const {return _originalSource;}

    /** Returns the identifier of the original target vertex. */
    NodeID originalTarget() const {return _originalTarget;}

    
 private:
    NodeID _originalSource;
    NodeID _originalTarget;
};


/**
   Creates a RelabeledEdgeWithoutSource from a RelabeledEdge.
   The source vertex is thrown away.
*/
RelabeledEdgeWithoutSource::RelabeledEdgeWithoutSource( const RelabeledEdge &edge )
    : EdgeWithoutSource( edge.target(), edge.weight() ),
      _originalSource( edge.originalSource() ), _originalTarget( edge.originalTarget() )
{}


#endif // RELABELEDEDGE_H
