/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                sparingStack.h                    *
 *                                                  *
 *  May 15, 2003                                    *
 ****************************************************/


#ifndef SPARINGSTACK_H
#define SPARINGSTACK_H

/**
   A CommonPoolOfBlocks manages blocks which can contain
   several elements of a specified value_type.
   It can be used by several SparingStacks.
*/
template <typename value_type, int elementsPerBlock, int noOfBlocks>
class CommonPoolOfBlocks
{
 public:
    /**
       A block which can contain several elements.
    */
    class Block
    {
     public:
	/** The default constructor. */
	Block()
	    : _size( 0 ), _prev( 0 )
	    {}

	/**
	   Adds a element to the block.
	   Precondition(!): full() must return false.
	*/
	void push(const value_type &element) {
	    _elements[_size++] = element;
	}

	/**
	   Returns a reference to the last element.
	   Precondition(!): empty() must return false.
	*/
	const value_type & top() const {return _elements[_size-1];}

	/**
	   Removes the last element.
	   Precondition(!): empty() must return false.
	*/
	void pop() {_size--;}

	/**
	   Removes all elements so that the block is empty.
	*/
	void clear() {_size = 0;}

	/** Returns true iff the block contains no elements. */
	bool empty() const {return (_size==0);}

	/** Returns true iff the block can't adopt more elements. */
	bool full() const {return (_size==elementsPerBlock);}

	/** Returns the number of elements which are stored in the block. */
	int size() const {return _size;}

	/** Returns a reference to the element which is stored at the given position. */
	const value_type & operator[](int index) const {return _elements[index];}

	/** Sets the pointer to the previous block. */
	void setPrevBlock(Block *const prev) {_prev = prev;}

	/** Returns a pointer to the previous block. */
	Block * prevBlock() const {return _prev;}
	
     private:
	/** The number of elements which are stored in the block. */
	int _size;

	/** A pointer to the previous block. */
	Block *_prev;

	/** The elements. */
	value_type _elements[elementsPerBlock];
    };
    

    /** The default constructor. */
    CommonPoolOfBlocks()
	: _free( &_blocks[noOfBlocks-1] ), _reserveMemory( 0 ),
	  _totalNoOfBlocks( noOfBlocks ), _noOfFreeBlocks( noOfBlocks )
	{
	    // initialize the linked list of free blocks
	    for (int i=noOfBlocks-1; i>0; i--)
		_blocks[i].setPrevBlock(&_blocks[i-1]);
	}

    /**
       The destructor.
       Blocks which have been created additionally are deleted.       
    */
    ~CommonPoolOfBlocks() {
	for (int i=0; i<_additionalBlocks.size(); i++) {
	    delete[] _additionalBlocks[i];
	    DS_VERBOSE1( std::cout << std::endl
			 << "CommonPoolOfBlocks::~CommonPoolOfBlocks(): "
			 << "additionally created blocks deleted." << std::endl );
	}
    }

    /** Returns a pointer to a free block. */
    Block * request() {
	
	if ( ! _free ) { // If there are no more free blocks...
	    // check if reserve memory is available
	    int noOfNewBlocks = _reserveMemory / sizeof( Block );
	    if (noOfNewBlocks == 0) {
		// If there is not enough reserve memory, the program is aborted !
		std::cerr << "CommonPoolOfBlocks: request(): No free blocks available !"
			  << std::endl;
		abort();
	    }

	    DS_VERBOSE1( std::cout << std::endl << "CommonPoolOfBlocks::request(): "
			 << noOfNewBlocks << " new blocks created." << std::endl );

	    // allocate new blocks according to the available reserve memory
	    _reserveMemory -= noOfNewBlocks * sizeof( Block );
	    
	    DS_LOGGING2( _totalNoOfBlocks += noOfNewBlocks;
			 _noOfFreeBlocks += noOfNewBlocks );

	    Block *newBlocks = new Block[noOfNewBlocks];
	    _additionalBlocks.push_back( newBlocks );
	    for (int i=0; i<noOfNewBlocks; i++) {
		newBlocks[i].setPrevBlock( _free );
		_free = &newBlocks[i];
	    }
	}
	
	DS_LOGGING2( _noOfFreeBlocks-- );

	// return a free block
	Block *block = _free;
	_free = _free->prevBlock();
	return block;
    }

    /** Adds a block which is no longer used to the free-list. */
    void release(Block *block) {
	block->setPrevBlock(_free);
	_free = block;

	DS_LOGGING2( _noOfFreeBlocks++ );
    }

    /**
       Informs that more reserve memory is available.
       @param newMemory the space in bytes which can be used by the
                        request method if more blocks are required.
    */
    void increaseReserveMemory(int newMemory) {
	_reserveMemory += newMemory;
    }

    
    DS_LOGGING2(
	/** Returns the number of used blocks. */
	int noOfBlocksUsed() const {return _totalNoOfBlocks - _noOfFreeBlocks;}
	
	/** Returns the number of free blocks. */
	int noOfBlocksFree() const {return _noOfFreeBlocks;}
    )
	
 private:
    /** The blocks which are initially created. */
    Block _blocks[noOfBlocks];

    /** A pointer to the linked list of free blocks. */
    Block *_free;

    /**
       The reserve memory in bytes which can be used by the
       request method if more blocks are required.
    */
    int _reserveMemory;

    std::vector<Block*> _additionalBlocks;
    
    /**
       The total number of blocks. This is the 'noOfBlocks' plus the
       number of blocks which are created additionally using the
       reserve memory.
    */
    int _totalNoOfBlocks;
    
    /** The number of free blocks. */
    int _noOfFreeBlocks;
    
};


/**
   A stack which stores its elements in blocks which are managed by
   a CommonPoolOfBlocks.
*/
template <typename value_type, int elementsPerBlock, int noOfBlocks>
class SparingStack
{
 protected:
    /** The type of the used CommonPoolOfBlocks. */
    typedef CommonPoolOfBlocks<value_type, elementsPerBlock, noOfBlocks> Pool;
    /** The type of a Block which stores several elements. */
    typedef typename Pool::Block Block;
//    typedef CommonPoolOfBlocks<value_type, elementsPerBlock, noOfBlocks>::Block Block;
    
 public:
    /** The default constructor. */
    SparingStack() : _top( &_bottom ) {}

    /** The destructor. */
    ~SparingStack() {
	while (_top != &_bottom) {
	    Block *oldBlock = _top;
	    _top = _top->prevBlock();
	    oldBlock->clear();
	    _pool->release(oldBlock);
	}
    }

    /**
       Sets the CommonPoolOfBlocks which should be used.
       This method must be called after the object has been created and
       before it is used.
    */
    void setPool(Pool *pool) {_pool = pool;}

    /** Returns true iff the stack is empty. */
    bool empty() const {return _top->empty();}

    /** Adds an element to the stack. */
    void push(const value_type &element) {
	if ( _top->full() ) {
	    // If the current block is full, request a new one.
	    Block *newBlock = _pool->request();
	    newBlock->setPrevBlock(_top);
	    _top = newBlock;
	}
	_top->push(element);
    }

    /** Returns a reference to the last element. */
    const value_type & top() const {return _top->top();}

    /** Removes the last element. */
    void pop() {
	_top->pop();
	if ( (_top->empty()) && (_top != &_bottom) ) {
	    // If the current block is now empty, release it.
	    // (The 'bottom' block belongs to the stack and is never released.)
	    Block *oldBlock = _top;
	    _top = _top->prevBlock();
	    _pool->release(oldBlock);
	}
    }

    /** Returns the number of elements which are stored in the stack. */
    int size() const {
	Block *current = _top;
	int blockCounter = 0;
	while (current != &_bottom) {
	    blockCounter++;
	    current = current->prevBlock();
	}
	return _top->size() + (blockCounter * DS_INT_EDGES_PER_BLOCK);
    }
    
 protected:
    /** A pointer to the used CommonPoolOfBlocks. */
    Pool *_pool;

    /** A pointer to the list of blocks which store the elements. */
    Block *_top;

    /** The bottom block which belongs to the stack. */
    Block _bottom;  
};


/**
   A specialized SparingStack which stores RelabeledEdgeWithoutSource-objects.
*/
template <int elementsPerBlock, int noOfBlocks>
class REWS_SparingStack : public SparingStack<RelabeledEdgeWithoutSource,
					      elementsPerBlock,
					      noOfBlocks>
{
 protected:
	typedef typename SparingStack<RelabeledEdgeWithoutSource, elementsPerBlock, noOfBlocks>::Block Block;

 public:
    /**
       Determines the edge with minimum weight.
       It is added to the resulting minimum spanning tree and
       the target node ID is returned.
       @param result a reference to the MST object which stores
                     the resulting minimum spanning tree
    */
    NodeID determineMinEdge(MST &result) const {
	// determine the edge with minimum weight
	RelabeledEdgeWithoutSource minEdge = this->top();
	Block *current = this->_top;
	while (true) {
	    for (int index=current->size()-1; index >= 0; index--) {
		if ( (*current)[index] < minEdge ) minEdge = (*current)[index]; 
	    }
	    if (current == &(this->_bottom)) break;
	    current = current->prevBlock();
	}

	// Add the edge with minimum weight incident to the
	// current node to the resulting minimum spanning tree.
	result.add( minEdge );
		
	DS_VERBOSE2( std::cout << "( !!! MST::add = "
		     << minEdge << " )" );

	return minEdge.target();
    }
};


#endif // SPARINGSTACK_H
