/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                pQueue.h                          *
 *                                                  *
 *  July 22, 2003                                   *
 ****************************************************/

#ifndef PQUEUE_H
#define PQUEUE_H


#include "../stxxl/containers/priority_queue.h"

#include "../data_structures/edgeVector.h"
#include "../data_structures/mst.h"


/**
   Represents an algorithm for reducing the number of nodes
   of a graph (in order to compute a minimum spanning tree)
   and the required data structures.
   This implementation uses an external priority queue.
*/
class PQueue
{
 private:
    /** The type of the priority queue that is used during the node reduction phase. */
    typedef stxxl::PRIORITY_QUEUE_GENERATOR<RelabeledEdge,
					    SourceWeightOrdering<RelabeledEdge>,
					    DS_PQUEUE_INTERNAL_MEMORY,
					    DS_PQUEUE_MAX_SIZE>::result PriorityQueue;
    
 public:
    /**
       Reduces the number of nodes of a graph (in order to
       compute a minimum spanning tree).
       @param graph a reference to an EdgeVector which represents the graph
       @param result a reference to a MST object which stores the result
       @param noOfNodesInIntMem the number of nodes which fit in internal memory
    */
    PQueue(EdgeVector<Edge> &graph, MST &result, NodeCount noOfNodesInIntMem)
	: _graph( graph ), _result( result ), 
	  _noOfNodesInIntMem( noOfNodesInIntMem ),
	  _prefetchPool( DS_PQUEUE_PREFETCH_POOL ), _writePool( DS_PQUEUE_WRITE_POOL ),
	  _pqueue(_prefetchPool, _writePool)
	{
	    initPQueue();
	    reduceNodes();
	}
    
    /**
       Returns a pointer to the first external bucket which contains the
       edges of the nodes which fit in internal memory.
    */
    InternalMemoryBucket* getIntMemBucket() {return _firstExtBucket;}

    /**
       Adds the given edge to the first external bucket or to the
       priority queue depending on the source vertex ID.
    */
    void add(const RelabeledEdge &edge);
    
 private:
    /** Reference to the given graph. */
    EdgeVector<Edge> &_graph;

    /** Reference to a MST object which stores the result. */
    MST &_result;

    /** The number of nodes which fit in internal memory. */
    NodeCount _noOfNodesInIntMem;

    /**
       The first external bucket which contains the
       edges of the nodes which fit in internal memory.
    */
    InternalMemoryBucket *_firstExtBucket;
    
    /** The prefetch pool that is used by the priority queue. */
    stxxl::prefetch_pool<PriorityQueue::block_type> _prefetchPool;

    /** The write pool that is used by the priority queue. */
    stxxl::write_pool<PriorityQueue::block_type> _writePool;

    /** The priority queue. */
    PriorityQueue _pqueue;
    
    /**
       Initializes the priority queue and distributes the edges to the
       first external bucket and the priority queue.
    */
    void initPQueue();

    /**
       Reduces the number of nodes.
       Only the first external bucket "survives".
    */
    void reduceNodes();
};


/**
   Adds the given edge to the first external bucket or to the
   priority queue depending on the source vertex ID.
*/
inline void PQueue::add(const RelabeledEdge &edge)
{
    if ( edge.isSelfLoop() ) return; // Self loops can be ignored.
    
    if (edge.source() < _noOfNodesInIntMem) {
	// add to the first external bucket
	_firstExtBucket->push_back( edge );
    }
    else {
	// add to the priority queue
	_pqueue.push( edge );
    }
}


#endif // PQUEUE_H
