/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                generate.cpp                      *
 *                                                  *
 *  April 13, 2003                                  *
 ****************************************************/


#include "generate.h"
#include "geometricGraphGenerator.h"

/**
 Generates a complete graph with the given number of nodes.
 The weights of the edges are chosen randomly.
*/
EdgeVector<> * generateCompleteGraph(NodeCount noOfNodes)
{
    EdgeCount noOfEdges = noOfNodes * (noOfNodes - 1) / 2;
    EdgeVector<> *el = new EdgeVector<>(noOfNodes, noOfEdges);

    for (NodeID i=0; i<noOfNodes; i++)
	for (NodeID j=0; j<i; j++) {
	    EdgeWeight weight = randomEdgeWeight();
	    Edge edge(i,j,weight);
	    el->push_back(edge);
	}
    
    return el;
}


/**
   Generates a random graph with the given number of nodes
   and the given number of edges.
   The weights of the edges are chosen randomly, too.
*/
EdgeVector<> * generateRandomGraph(NodeCount noOfNodes,
				 EdgeCount noOfEdges)
{
    EdgeVector<> *el = new EdgeVector<>(noOfNodes, noOfEdges);

    DS_VERBOSE1( std::cout << std::endl << "generate random graph" << std::endl;
		 Percent percent(noOfEdges) );
    for (EdgeCount i=0; i<noOfEdges; i++) {
	DS_VERBOSE1( percent.printStatus(i) );
	NodeID source = randomNodeID( noOfNodes );
	NodeID target = randomNodeID( noOfNodes );
	EdgeWeight weight = randomEdgeWeight();
	Edge edge(source,target,weight);
	el->push_back(edge);
    }
    
    return el;  
}


/**
   Generates a grid graph with "noOfNodesX * noOfNodesY" nodes.
   The weights of the edges are chosen randomly.
*/
EdgeVector<> * generateGridGraph(NodeCount noOfNodesX, NodeCount noOfNodesY)
{
    EdgeCount noOfEdges = 2 * noOfNodesX * noOfNodesY - noOfNodesX - noOfNodesY; 
    EdgeVector<> *el = new EdgeVector<>(noOfNodesX * noOfNodesY, noOfEdges);

    DS_VERBOSE1( std::cout << std::endl << "generate grid graph" << std::endl;
		 Percent percent(noOfEdges) );
    EdgeCount i = 0;
    for (NodeCount x=0; x<noOfNodesX; x++)
	for (NodeCount y=0; y<noOfNodesY; y++) {
	    NodeID source = (x * noOfNodesY) + y;
	    if ( x > 0 ) {
		DS_VERBOSE1( percent.printStatus(i++) );
		
		NodeID target = ((x-1) * noOfNodesY) + y;
		EdgeWeight weight = randomEdgeWeight();
		Edge edge(source,target,weight);
		el->push_back(edge);
	    }
	    if ( y > 0 ) {
		DS_VERBOSE1( percent.printStatus(i++) );
		
		NodeID target = (x * noOfNodesY) + (y-1);
		EdgeWeight weight = randomEdgeWeight();
		Edge edge(source,target,weight);
		el->push_back(edge);
	    }
	}
    
    return el;  
}


/**
   Generates a geometric graph.
   The given number of nodes is placed in a square and each node is connected with
   the given number of nearest neighbours. Duplicates (which come into existence when
   two nodes select each other) are removed.
*/
EdgeVector<> * generateGeometricGraph(NodeCount noOfNodes, NodeCount noOfNeighbours)
{
    GeometricGraphGenerator *generator = new GeometricGraphGenerator(noOfNodes, noOfNeighbours);
    EdgeVector<> *generatedGraph = generator->generatedGraph();
    delete generator;
    return generatedGraph;
}

/** Returns a random node identifier between 0 and n-1. */
inline NodeID randomNodeID(NodeID n) {
    return (NodeID)(rand() / (double)(RAND_MAX+1.0) * n);
}

/** Returns a random edge weight. */
inline EdgeWeight randomEdgeWeight() {
    const EdgeWeight maxEdgeWeight = 1e6;
    return (EdgeWeight)(rand() / (double)(RAND_MAX+1.0) * maxEdgeWeight);
}
