/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                geometricGraphGenerator.h         *
 *                                                  *
 *  June 09, 2003                                   *
 ****************************************************/


#ifndef GEOMETRICGRAPHGENERATOR_H
#define GEOMETRICGRAPHGENERATOR_H

#include "generate.h"
#include "../data_structures/edgeVector.h"
#include "../data_structures/sparingStack.h"
#include "../stxxl/containers/vector.h"
#include "../stxxl/algo/sort.h"


/**
   Generates a geometric graph.
   The given number of nodes is placed in a square and each node is connected with
   the given number of nearest neighbours. Duplicates (which come into existence when
   two nodes select each other) are removed.
   The distance between two nodes is the square of the Euclidean distance (the extraction
   of the root is unnecessary).
*/
class GeometricGraphGenerator
{
 private:
    /** number of dots per dimension */
    static const int _resolution = 1e6;

    /** number of big squares per dimension */
    static const int _noOfBigSquares = 20;

    /** number of active big squares per dimension */
    static const int _noOfActiveBigSquares = 3;

    /** number of little squares per dimension */
    static const int _noOfLittleSquares = 400;

    /** number of dots per big square per dimension */
    static const int _dotsPerBigSquare = _resolution / _noOfBigSquares;

    /** number of dots per little square per dimension */
    static const int _dotsPerLittleSquare = _dotsPerBigSquare / _noOfLittleSquares;

    /** number of elements per block in the common pool used by the active big squares */
    static const int _poolElementsPerBlock = 10;

    /** number of blocks in the common pool used by the active big squares */
    static const int _poolNoOfBlocks = 3e6;

    /** the default block size divided by this constant is used as block size for external vectors */
    static const int _externalBlockSizeDivisor = 32;


    /** Represents a x- and a y-coordinate. */
    class Coordinates
    {
	friend std::ostream& operator<<( std::ostream& os, const Coordinates &coord ) {
	    os << coord._x << " / " << coord._y;
	    return os;
	}
     public:
	Coordinates() : _x( 0 ), _y( 0 ) {}
	Coordinates(int x, int y) : _x( x ), _y( y ) {}
	int x() const {return _x;}
	int y() const {return _y;}
	void x(int x) {_x = x;}
	void y(int y) {_y = y;}
	void incX() {_x++;}
	void incY() {_y++;}
	void decX() {_x--;}
	void decY() {_y--;}
	void divideBy(int divisor) {_x /= divisor; _y /= divisor;}
	void modulo(int divisor) {_x %= divisor; _y %= divisor;}
	void multiply(int factor) {_x *= factor; _y *= factor;}
	void add(Coordinates addend) {_x += addend.x(); _y += addend.y();}
	void add(int addend) {_x += addend; _y += addend;} 
	void subtract(Coordinates subtrahend) {_x -= subtrahend.x(); _y -= subtrahend.y();}
	int min() {return (_x < _y) ? _x : _y;}
	int max() {return (_x > _y) ? _x : _y;}
	bool inRange(int lowerBound, int upperBound) {
	    return (_x >= lowerBound) && (_x < upperBound) && (_y >= lowerBound) && (_y < upperBound);
	}
     private:
	int _x, _y;
    };

    /** Represents a node in the geometric graph. */
    class Node
    {
     public:
	Node() {}
	Node(NodeID id, Coordinates pos) : _id( id ), _pos( pos ) {}

	/** Returns the coordinates of the node. */
	const Coordinates & pos() const {return _pos;}

	/** Returns the id of the node. */
	NodeID id() const {return _id;}

     private:
	NodeID _id;
	Coordinates _pos;
    };

    /**
       A subclass of "SparingStack" that provides iterator-methods.
    */
    template <typename value_type, int elementsPerBlock, int noOfBlocks>
    class SparingStackWithIterator :
	public SparingStack<value_type,	elementsPerBlock, noOfBlocks>
    {
     protected:
	typedef typename SparingStack<value_type, elementsPerBlock, noOfBlocks>::Block Block;

     public:
	/**
	   Returns the first element of the stack and (re)initializes the iterator.
	   Precondition: empty() returns false.
	*/
	const value_type & first() {
	    _currentBlock = this->_top;
	    _currentElement = _currentBlock->size() - 1;
	    return (*_currentBlock)[_currentElement];
	}

	/**
	   Returns the next element of the stack.
	   Precondition: hasMoreElements() returns true.
	*/
	const value_type & next() {
	    if (_currentElement == 0) {
		_currentBlock = _currentBlock->prevBlock();
		_currentElement = _currentBlock->size();
	    }
	    _currentElement--;
	    return (*_currentBlock)[_currentElement];
	}
	
	/** Returns true iff the next call of "next()" will be successful. */
	bool hasMoreElements() {
	    return (_currentElement > 0) || (_currentBlock != &(this->_bottom));
	}
	
     private:
	Block *_currentBlock;
	int _currentElement;
    };

    /** Represents a big square that is loaded in internal memory. */
    class BigSquare;

    
 public:
    /**
       Generates a geometric graph.
       The given number of nodes is placed in a square and each node is connected with
       the given number of nearest neighbours. Duplicates (which come into existence when
       two nodes select each other) are removed.
    */
    GeometricGraphGenerator(NodeCount noOfNodes, NodeCount noOfNeighbours)
	: _edges( new EdgeVector<>(noOfNodes,0) ), _pool( new Pool() ),
	  _active( _noOfBigSquares, _noOfBigSquares )
	{
	    DS_VERBOSE1( std::cout << std::endl << "generate geometric graph" << std::endl );

	    generateNodes(noOfNodes);
	    generateEdges(noOfNeighbours);
	    removeDuplicates();
	}

    /** Returns a pointer to the generated graph. */
    EdgeVector<> * generatedGraph() const {return _edges;}

    
 private:
    /** The type of the CommonPoolOfBlocks that is used by the active big squares. */
    typedef CommonPoolOfBlocks<Node,_poolElementsPerBlock,_poolNoOfBlocks> Pool;

    /** The type of the SparingStack that is used by the active big squares. */
    typedef SparingStackWithIterator<Node,_poolElementsPerBlock,_poolNoOfBlocks> SpSt;

    /** The big squares in external memory. */
    stxxl::VECTOR_GENERATOR<Node,
			    DS_DEFAULT_PAGE_SIZE,
			    DS_DEFAULT_NO_OF_PAGES,
			    DS_DEFAULT_BLOCK_SIZE / _externalBlockSizeDivisor>::result
        _bigSquares[_noOfBigSquares][_noOfBigSquares];

    /** The active big squares in internal memory. */
    BigSquare *_activeBigSquares[_noOfActiveBigSquares][_noOfActiveBigSquares];

    /** The coordinates of the left upper corner of the active big squares. */
    Coordinates _active;

    /** The generated edges. */
    EdgeVector<> *_edges;

    /** The common pool used by the active big squares. */
    Pool *_pool;


    /** Generates the nodes. */
    void generateNodes(NodeCount noOfNodes) {
	DS_VERBOSE1( std::cout << "  generate nodes" << std::endl;
		     Percent percent(noOfNodes) );
	for (NodeCount i=0; i<noOfNodes; i++) {
	    DS_VERBOSE1( percent.printStatus(i) );
	    
	    Node node( i, Coordinates((int)randomNodeID(_resolution),
				      (int)randomNodeID(_resolution)) );
	    push(node);
	    
	    DS_VERBOSE2( std::cout << "node " << node.id() << " (" << node.pos() << ")" << std::endl );
	}
    }

    /** Adds a generated node to the appropriate big square. */
    void push(const Node &node) {
	Coordinates b = node.pos();
	b.divideBy( _dotsPerBigSquare );
	_bigSquares[b.x()][b.y()].push_back(node);
    }

    /** Generates the edges. */
    void generateEdges(NodeCount noOfNeighbours) {
	DS_VERBOSE1( std::cout << std::endl << "  generate edges" << std::endl;
		     Percent percent(_noOfBigSquares * _noOfBigSquares) );

	// Process all big squares and for each big square all little squares
	// and for each little square all nodes
	Coordinates b,l;
	for (b.x(0); b.x()<_noOfBigSquares; b.incX())
	    for (b.y(0); b.y()<_noOfBigSquares; b.incY()) {

		DS_VERBOSE1( percent.printStatus(b.x() * _noOfBigSquares + b.y()) );
		
		for (l.x(0); l.x()<_noOfLittleSquares; l.incX())
		    for (l.y(0); l.y()<_noOfLittleSquares; l.incY()) {
			SpSt *currentLiSq = littleSquare(b,l);
			if ( ! currentLiSq->empty() ) {
			    std::vector<Node> tmp;
			    tmp.reserve(currentLiSq->size());
			    Node node = currentLiSq->first();
			    // copy all nodes of the current little square to a temporary vector
			    // (because the call of first() in "findNeighbours(...)" would confuse
			    // the iterator)
			    while (true) {
				tmp.push_back(node);
				if ( ! currentLiSq->hasMoreElements() ) break;
				node = currentLiSq->next();
			    }

			    // process all nodes in the temporary vector
			    do {
				Node node = tmp.back();
				DS_VERBOSE2( std::cout << "find neighbours for node " << node.id() << " in "
					     << b << " - " << l << ": " );
				int edgesToDo = noOfNeighbours;
				int radiusUnits = distanceFromBorder(b, l, node.pos());
				int radiusLittleSq = 0;
				// iterate until enough neighbours are found
				// first, only the current little square is searched
				// then, additionally the 8 surrounding squares are searched
				// and so on
				while (edgesToDo > 0) {
				    DS_VERBOSE2( std::cout << edgesToDo << " "; std::cout.flush() );
				    edgesToDo -= findNeighbours(b, l, node, radiusUnits, radiusLittleSq, edgesToDo);
				    radiusUnits += _dotsPerLittleSquare;
				    radiusLittleSq++;
				}
				DS_VERBOSE2( std::cout << std::endl );
				tmp.pop_back();
			    } while ( ! tmp.empty() );
			}
		    }
	    }
    }

    /**
       Returns the shortest distance of the node whose coordinates are given to one of the four
       borders of the given little square in the given big square.
    */
    int distanceFromBorder(const Coordinates b, const Coordinates l, const Coordinates pos) const {
	Coordinates left_top( b.x() * _dotsPerBigSquare + l.x() * _dotsPerLittleSquare,
			      b.y() * _dotsPerBigSquare + l.y() * _dotsPerLittleSquare );
	Coordinates right_bottom = left_top;
	right_bottom.add(_dotsPerLittleSquare-1);
	left_top.subtract(pos);
	left_top.multiply(-1);
	right_bottom.subtract(pos);
	return (left_top.min() < right_bottom.min()) ? left_top.min() : right_bottom.min();
    }

    /**
       Searches the given little square (in the given big square) and (dependent on 'radiusLittleSq')
       the surrounding little squares in order to find the nearest 'edgesToDo' neighbours for the
       given node. The distance of regarded neighbours is limited by 'radiusUnits'.
    */
    int findNeighbours(const Coordinates b, const Coordinates l_abs, const Node node,
		       int radiusUnits, int radiusLittleSq, int edgesToDo) {
	// There is a lower and a upper bound for the distance. The lower bound prevents that
	// neighbours are found which already have been found. The upper bound ensures that
	// only neighbours are regarded that lie in a radius that is covered by the searched
	// little squares.
	int distanceLowerBound = -1;
	if (radiusLittleSq > 0)
	    distanceLowerBound = (radiusUnits - _dotsPerLittleSquare) * (radiusUnits - _dotsPerLittleSquare);
	int distanceUpperBound = radiusUnits * radiusUnits;

	// contains the nearest neighbours for the current node
	std::pair<Node,int> nearestNeighbours[edgesToDo];	
	int noOfNeighbours = 0;

	// search the little squares
	Coordinates l_rel;
	for (l_rel.x(-radiusLittleSq); l_rel.x() <= radiusLittleSq; l_rel.incX())
	    for (l_rel.y(-radiusLittleSq); l_rel.y() <= radiusLittleSq; l_rel.incY()) {
		Coordinates l = l_abs;
		l.add(l_rel);
		SpSt *currentLiSq = littleSquare(b,l);		
		if ( (currentLiSq) && (! currentLiSq->empty()) ) {
		    // process all nodes in the current little square
		    Node neighbour = currentLiSq->first();
		    while (true) {
			if (node.id() != neighbour.id()) {
			    // compute the distance
			    int distance = getDistance(node.pos(), neighbour.pos());
			    // check the lower and upper bound for the distance
			    if ((distance > distanceLowerBound) && (distance <= distanceUpperBound)) {
				// insert the current neighbour in the array that contains
				// the nearest neighbours
				int i = noOfNeighbours;
				while (i > 0) {
				    if (distance < nearestNeighbours[i-1].second) {
					if (i < edgesToDo) nearestNeighbours[i] = nearestNeighbours[i-1];
					i--;
				    }
				    else {
					break;
				    }
				}
				if (i < edgesToDo) {
				    nearestNeighbours[i] = std::pair<Node,int>(neighbour,distance);
				    if (noOfNeighbours < edgesToDo) noOfNeighbours++;
				}
			    }
			}
			if ( ! currentLiSq->hasMoreElements() ) break;
			neighbour = currentLiSq->next();
		    }
		}
	    }
	// create edges from the current node to the nearest neighbours that have been found
	for (int i=0; i<noOfNeighbours; i++) {
	    Edge edge(node.id(),nearestNeighbours[i].first.id(),nearestNeighbours[i].second);
	    _edges->push_back(edge);
	}
	// return the number of neighbours that have been found
	return noOfNeighbours;
    }

    /**
       Returns the distance between two points.
       The distance between two points is the square of the Euclidean distance (the extraction
       of the root is unnecessary). 
    */
    int getDistance(const Coordinates a, const Coordinates b) {
	return (a.x() - b.x())*(a.x() - b.x()) + (a.y() - b.y())*(a.y() - b.y());
    }

    /**
       Returns a pointer to the requested little square in the given big square.
       If the big square is active, i.e. it is loaded in internal memory, the pointer
       is returned at once. Otherwise, the required big square is loaded and then
       the pointer is returned.
       If the requested little square doesn't exist, 0 is returned.
    */
    SpSt * littleSquare(Coordinates b, Coordinates l) {
	// correct little square coordinates that are out of range
	// by adapting the big square coordinates
	while (l.x() < 0) {l.x( l.x() + _noOfLittleSquares ); b.decX();}
	while (l.y() < 0) {l.y( l.y() + _noOfLittleSquares ); b.decY();}
	while (l.x() >= _noOfLittleSquares) {l.x( l.x() - _noOfLittleSquares ); b.incX();}
	while (l.y() >= _noOfLittleSquares) {l.y( l.y() - _noOfLittleSquares ); b.incY();}

	// return 0 if the requested little square doesn't exist
	if (b.min() < 0) return 0;
	if (b.max() >= _noOfBigSquares) return 0;
		
	b.subtract(_active);
	// check if the required big square is loaded
	if ( ! b.inRange( 0, _noOfActiveBigSquares ) ) {
	    // The required big square isn't loaded. Hence, shift the frame of active
	    // big squares.
	    Coordinates shift;
	    if (b.x() < 0) shift.x( b.x() );
	    if (b.x() >= _noOfActiveBigSquares) shift.x( b.x() - (_noOfActiveBigSquares-1) );
	    if (b.y() < 0) shift.y( b.y() );
	    if (b.y() >= _noOfActiveBigSquares) shift.y( b.y() - (_noOfActiveBigSquares-1) );

	    DS_VERBOSE2( std::cout << "active big squares shift: " << shift << std::endl );

	    // shift loaded big squares to the new position in the frame
	    BigSquare *tmpBigSquares[_noOfActiveBigSquares][_noOfActiveBigSquares];
	    Coordinates i;
	    for (i.x(0); i.x()<_noOfActiveBigSquares; i.incX())
		for (i.y(0); i.y()<_noOfActiveBigSquares; i.incY()) {
		    Coordinates active = i;
		    active.add( shift );
		    if ( active.inRange(0, _noOfActiveBigSquares) ) {
			tmpBigSquares[i.x()][i.y()] = _activeBigSquares[active.x()][active.y()];
			_activeBigSquares[active.x()][active.y()] = 0;
		    }
		}
	    // delete active big squares that have been ruled out by the shifting
	    for (i.x(0); i.x()<_noOfActiveBigSquares; i.incX())
		for (i.y(0); i.y()<_noOfActiveBigSquares; i.incY()) {
		    delete _activeBigSquares[i.x()][i.y()];
		    _activeBigSquares[i.x()][i.y()] = tmpBigSquares[i.x()][i.y()];
		}
	    // load big squares to fill the free spaces in the frame
	    for (i.x(0); i.x()<_noOfActiveBigSquares; i.incX())
		for (i.y(0); i.y()<_noOfActiveBigSquares; i.incY()) {
		    Coordinates active = i;
		    active.add( shift );
		    if ( ! active.inRange(0, _noOfActiveBigSquares) ) {
			Coordinates load = active;
			load.add( _active );
			if ( load.inRange(0, _noOfBigSquares) ) {

			    DS_VERBOSE2( std::cout << "load big square: " << load
					 << "; size = " << _bigSquares[load.x()][load.y()].size() << std::endl );

			    // load all nodes from the current big square and distribute them
			    // to the appropriate little squares
			    _activeBigSquares[i.x()][i.y()] = new BigSquare(_pool);
			    for (int j=0; j<_bigSquares[load.x()][load.y()].size(); j++) {
				Node node = _bigSquares[load.x()][load.y()][j];
				Coordinates posInLittleSq = node.pos();
				posInLittleSq.modulo(_dotsPerBigSquare);
				posInLittleSq.divideBy(_dotsPerLittleSquare);
				(_activeBigSquares[i.x()][i.y()]->littleSquare(posInLittleSq))
				    ->push(node);
			    }
			}
		    }
		}
	    _active.add(shift);
	    b.subtract(shift);
	}	
	return _activeBigSquares[b.x()][b.y()]->littleSquare(l);
    }

    /** Removes duplicates, which come into existence when two nodes select each other */
    void removeDuplicates() {
	DS_VERBOSE1( std::cout << std::endl << "  remove duplicates" << std::endl;
		     std::cout << "    (ensure source > target)" << std::endl;
	             system("date") );

	// delete the active big squares and the common pool
	Coordinates i;
	for (i.x(0); i.x()<_noOfActiveBigSquares; i.incX())
	    for (i.y(0); i.y()<_noOfActiveBigSquares; i.incY()) 
		delete _activeBigSquares[i.x()][i.y()];
	delete _pool;
	
	if (_edges->noOfEdges() == 0) return;

	// process all edges and ensure that source > target
	for (EdgeCount i=0; i<_edges->noOfEdges(); i++) {
	    Edge &edge = (*_edges)[i];
	    if (edge.source() < edge.target()) edge.swap();
	}

	DS_VERBOSE1( std::cout << "    (sort)" << std::endl;
	             system("date") );
	
	// sort the edges
	stxxl::sort(_edges->begin(),_edges->end(),
		    SourceTargetWeightOrdering(),DS_INTERNAL_MEMORY_FOR_SORTING - 100*1024*1024);

	DS_VERBOSE1( std::cout << "    (scan)" << std::endl;
	             system("date") );
	
	// scan the sorted edges and add an edge to the generated graph if it differs
	// from the predecessor
	EdgeVector<> *edgesWithoutDupl = new EdgeVector<>(_edges->noOfNodes(),0);
	edgesWithoutDupl->push_back( _edges->back() );
	_edges->pop_back();
	while ( ! _edges->empty() ) {
	    if ( ! (edgesWithoutDupl->back() == _edges->back()) )
		edgesWithoutDupl->push_back( _edges->back() );
	    _edges->pop_back();
	}
	delete _edges;
	_edges = edgesWithoutDupl;

	DS_VERBOSE1( std::cout << "    (finished)" << std::endl;
	             system("date") );
    }

    /** Represents a big square that is loaded in internal memory. */
    class BigSquare
    {
     public:
	/** The constructor. */
	BigSquare(Pool *pool) {
	    for (int x=0; x<_noOfLittleSquares; x++)
		for (int y=0; y<_noOfLittleSquares; y++)
		    littleSquares[x][y].setPool(pool);
	}

	/** Returns a pointer to the requested little square. */
	SpSt * littleSquare(Coordinates coord) {return &littleSquares[coord.x()][coord.y()];}
	
     private:
	/** The little squares. */
	SpSt littleSquares[_noOfLittleSquares][_noOfLittleSquares];
    };

    	    
};

#endif // GEOMETRICGRAPHGENERATOR_H
