/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                importExport.cpp                  *
 *                                                  *
 *  April 13, 2003                                  *
 ****************************************************/


#include "../stxxl/algo/sort.h"

#include "importExport.h"


/**
   Imports an EdgeVector from a file.
   expected format:
   the format which is expected by 'importEdgeVectorListOfEdges' or
   the format which is expected by 'importEdgeVectorAdjList' or
   the format which is expected by 'importEdgeVectorCompressed'
*/
EdgeVector<> * importEdgeVector(const std::string &filename)
{
    std::ifstream in( filename.c_str() );
    if (in.peek() == 'a') {
	return importEdgeVectorAdjList( in );
    }
    if (in.peek() == 'c') {
	in.close();
	return importEdgeVectorCompressed( filename );
    }
    return importEdgeVectorListOfEdges( in );
}


/**
   Imports an EdgeVector from an istream.
   expected format:
   
   noOfNodes noOfEdges
   source0 target0 weight0
   source1 target1 weight1
   source2 target2 weight2
   ...

   sourceX and targetX are node indizes; the first used node index should be 0.
*/
EdgeVector<> * importEdgeVectorListOfEdges(std::istream &in)
{
    NodeCount noOfNodes;
    EdgeCount noOfEdges;

    in >> noOfNodes >> noOfEdges;

    EdgeVector<> *el = new EdgeVector<>(noOfNodes,noOfEdges);
    
    for (EdgeCount i=0; i<noOfEdges; i++) {
	NodeID source,target;
	EdgeWeight weight;
	
	in >> source >> target >> weight;
	
	Edge edge(source,target,weight);
	el->push_back(edge);
    }
    
    return el;
}


/**
   Imports an EdgeVector from an istream.
   expected format:
   
   'a'
   noOfNodes
   adjacency list of node 1
   -1
   adjacency list of node 2
   -1
   ...

   Each adjacency list is a list of pairs: node1 weight1 node2 weight2 ...
   The first used node index should be 1.
*/
EdgeVector<> * importEdgeVectorAdjList(std::istream &in)
{
    NodeCount noOfNodes;
    
    char dummy;
    in >> dummy >> noOfNodes;

    EdgeVector<> *el = new EdgeVector<>(noOfNodes,0);

    for (NodeCount i=0; i<noOfNodes; i++) {
	NodeID target;
	EdgeWeight weight;
	
	in >> target;
	// target is decremented so that the node indizes start from 0
	while (target-- != -1) { 
	    in >> weight;

	    // iff the source index is greater than the target index,
	    // the edge is added to the list
	    if (i>target) {
		Edge edge(i,target,weight);		
		el->push_back(edge);
	    }
	    
	    in >> target;
	}
    }    

    return el;
}


/**
   Imports an EdgeVector from a file.
   expected format:

   'c'
   noOfNodes noOfEdges
   source0 target0 weight0
   source1 target1 weight1
   source2 target2 weight2
   ...

   sourceX and targetX are node indizes; the first used node index should be 0.
   Each 32-bit number (sourceX, targetX, weightX) should be represented by 4 chars
   and there should be no spaces and newlines after the number of edges.
   If necessary, several files are used so that the file size limit isn't exceeded.
*/
EdgeVector<> * importEdgeVectorCompressed(const std::string &filename)
{
    static const EdgeCount edgesPerFile = 150000000;
    EdgeCount edgesInCurrentFile = 0;
    char currentFileSuffix = 'a';
    
    std::ifstream in( filename.c_str() );
    
    NodeCount noOfNodes;
    EdgeCount noOfEdges;

    DS_VERBOSE1( std::cout << "Import started. " << std::flush; system("date") );
    
    char dummy;
    in >> dummy >> noOfNodes >> noOfEdges >> std::noskipws >> dummy;
    
    EdgeVector<> *el = new EdgeVector<>(noOfNodes,noOfEdges);
    
    for (EdgeCount i=0; i<noOfEdges; i++) {
	if (edgesInCurrentFile == edgesPerFile) {
	    DS_VERBOSE1(
		std::cout << "File " << currentFileSuffix << " completed. Altogether "
		<< i << " edges read. " << std::flush;
		system("date");
	    )
	    edgesInCurrentFile = 0;
	    currentFileSuffix++;
	    in.close();
	    in.open( (filename + "." + currentFileSuffix).c_str() );
	}
	
	unsigned int dataToRead[3];
	for (int j=0; j<3; j++) {
	    dataToRead[j] = 0;
	    int factor = 1;
	    for (int k=0; k<4; k++) {
		unsigned char tmp;
		in >> tmp;
		dataToRead[j] += tmp * factor;
		factor *= 256;
	    }
	}
	
	Edge edge(dataToRead[0],dataToRead[1],dataToRead[2]);
	el->push_back(edge);

	edgesInCurrentFile++;
    }
    DS_VERBOSE1(
	std::cout << "File " << currentFileSuffix << " completed. Altogether "
	          << noOfEdges << " edges read. " << std::flush;
	system("date");
    )
    return el;
}


/**
   Exports an EdgeVector to an ostream.
   created format:

   'a'
   noOfNodes
   adjacency list of node 1
   -1
   adjacency list of node 2
   -1
   ...

   Each adjacency list is a list of pairs: node1 weight1 node2 weight2 ...
   The first used node index is 1.
*/
void exportEdgeVectorAdjList(std::ostream &out, EdgeVector<> &edges)
{
    EdgeVector<> doubledEdges( edges.noOfNodes(), 2 * edges.noOfEdges() );
    
    for (EdgeCount i=0; i<edges.noOfEdges(); i++) {
	doubledEdges.push_back(edges[i]);
	doubledEdges.push_back(edges[i]);
	doubledEdges.back().swap();
    }

    stxxl::sort(doubledEdges.begin(),doubledEdges.end(),
		SourceTargetWeightOrdering(),DS_INTERNAL_MEMORY_FOR_SORTING);

    out << "a" << std::endl << edges.noOfNodes() << std::endl;
    EdgeCount j=0;
    for (NodeCount i=0; i<edges.noOfNodes(); i++) {
	for ( ; (j<doubledEdges.size()) && (doubledEdges[j].source()==i); j++ ) {
	    out << (doubledEdges[j].target()+1) << " " << doubledEdges[j].weight() << " ";
	}
	out << std::endl << "-1" << std::endl;
    }
}


/**
   Exports an EdgeVector to a file.
   created format:

   'c'
   noOfNodes noOfEdges
   source0 target0 weight0
   source1 target1 weight1
   source2 target2 weight2
   ...

   sourceX and targetX are node indizes; the first used node index is 0.
   Each 32-bit number (sourceX, targetX, weightX) is represented by 4 chars
   and there are no spaces and newlines after the number of edges.
   If necessary, several files are used so that the file size limit isn't exceeded.
*/
void exportEdgeVectorCompressed(const std::string &filename, EdgeVector<> &edges)
{
    static const EdgeCount edgesPerFile = 150000000;
    EdgeCount edgesInCurrentFile = 0;
    char currentFileSuffix = 'a';
    
    std::ofstream out( filename.c_str() );

    DS_VERBOSE1( std::cout << "Export started. " << std::flush; system("date") );
    out << 'c' << std::endl << edges.noOfNodes() << " " << edges.noOfEdges() << std::endl;
    for (EdgeCount i=0; i<edges.noOfEdges(); i++) {
	if (edgesInCurrentFile == edgesPerFile) {
	    DS_VERBOSE1(
		std::cout << "File " << currentFileSuffix << " completed. Altogether "
                          << i << " edges written. " << std::flush;
		system("date");
	    )
	    edgesInCurrentFile = 0;
	    currentFileSuffix++;
	    out << std::endl;
	    out.close();
	    out.open( (filename + "." + currentFileSuffix).c_str() );
	}
	
	unsigned int dataToWrite[] = {edges[i].source(), edges[i].target(), edges[i].weight()};
	for (int j=0; j<3; j++) {
	    for (int k=0; k<4; k++) {
		char tmp = dataToWrite[j] % 256;
		dataToWrite[j] /= 256;
		out << tmp;
	    }
	}
	
	edgesInCurrentFile++;
    }
    DS_VERBOSE1(
	std::cout << "File " << currentFileSuffix << " completed. Altogether "
	          << edges.noOfEdges() << " edges written. " << std::flush;
	system("date");
    )
    out << std::endl;
    out.close();
}


/**
   Returns the number of nodes of the graph which is stored in the file
   with the given name.
*/
NodeCount noOfNodesInFile(const std::string &filename)
{
    std::ifstream inFile(filename.c_str());
    if ((inFile.peek() == 'a') || (inFile.peek() == 'c')) {
	char dummy;
	inFile >> dummy;
    }
    NodeCount noOfNodes;
    inFile >> noOfNodes;
    return noOfNodes;
}
