/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                logging.h                         *
 *                                                  *
 *  May 11, 2003                                    *
 ****************************************************/

#ifndef LOGGING_H
#define LOGGING_H

#include "../stxxl/common/utils.h"
#include "../buckets/buckets.h"

#include <fstream>
#include <iomanip>

class LoggingProblemInstance
{
 public:
    LoggingProblemInstance(NodeCount noOfNodesInInternalMemory)
	: _noOfNodesInInternalMemory( noOfNodesInInternalMemory )
	{}
    
    virtual void printProblemInstance(std::ostream &out) const = 0;

    virtual EdgeCount noOfEdges() const = 0;
    virtual NodeCount noOfNodes() const = 0;
    
    NodeCount noOfNodesInInternalMemory() const {
	return _noOfNodesInInternalMemory;
    }
    
 protected:
    NodeCount _noOfNodesInInternalMemory;
};

class LoggingRandomGraph : public LoggingProblemInstance
{
 public:
    LoggingRandomGraph(NodeCount noOfNodes, EdgeCount noOfEdges,
		       NodeCount noOfNodesInInternalMemory)
	: LoggingProblemInstance(noOfNodesInInternalMemory),
	  _noOfNodes( noOfNodes ), _noOfEdges( noOfEdges )
	{}
    
    virtual void printProblemInstance(std::ostream &out) const {
	out << std::endl << "RANDOM GRAPH" << std::endl << std::endl
	    << std::scientific << std::setprecision(2)
	    << "#nodes = " << (double)_noOfNodes << std::endl
	    << "#edges = " << (double)_noOfEdges << std::endl
	    << "#nodes which fit in internal memory = " << (double)_noOfNodesInInternalMemory << std::endl
	    << "--------------------" << std::endl << std::fixed;
    }

    virtual EdgeCount noOfEdges() const {return _noOfEdges;}
    virtual NodeCount noOfNodes() const {return _noOfNodes;}
    
 private:
    NodeCount _noOfNodes;
    EdgeCount _noOfEdges;
};

class LoggingGridGraph : public LoggingProblemInstance
{
 public:
    LoggingGridGraph(NodeCount noOfNodesX, NodeCount noOfNodesY,
		     NodeCount noOfNodesInInternalMemory)
	: LoggingProblemInstance(noOfNodesInInternalMemory),
	  _noOfNodesX( noOfNodesX ), _noOfNodesY( noOfNodesY )
	{
	    _noOfEdges = 2 * noOfNodesX * noOfNodesY - noOfNodesX - noOfNodesY;
	}
    
    virtual void printProblemInstance(std::ostream &out) const {
	out << std::endl << "GRID GRAPH" << std::endl << std::endl
	    << std::scientific << std::setprecision(2)
	    << "#nodes = " << (double)(_noOfNodesX * _noOfNodesY)
	    << " ( = " << _noOfNodesX << " x " << _noOfNodesY << " )" << std::endl
	    << "#edges = " << (double)_noOfEdges << std::endl
	    << "#nodes which fit in internal memory = "
	    << (double)_noOfNodesInInternalMemory << std::endl
	    << "--------------------" << std::endl << std::fixed;
    }

    virtual EdgeCount noOfEdges() const {return _noOfEdges;}
    virtual NodeCount noOfNodes() const {return _noOfNodesX * _noOfNodesY;}
    
 private:
    NodeCount _noOfNodesX;
    NodeCount _noOfNodesY;
    EdgeCount _noOfEdges;
};

class LoggingGeometricGraph : public LoggingProblemInstance
{
 public:
    LoggingGeometricGraph(NodeCount noOfNodes, NodeCount noOfNeighbours,
			  EdgeCount noOfEdges,
			  NodeCount noOfNodesInInternalMemory)
	: LoggingProblemInstance(noOfNodesInInternalMemory),
	  _noOfNodes( noOfNodes ), _noOfNeighbours( noOfNeighbours ),
          _noOfEdges( noOfEdges )
	{}
    
    virtual void printProblemInstance(std::ostream &out) const {
	out << std::endl << "GEOMETRIC GRAPH" << std::endl << std::endl
	    << std::scientific << std::setprecision(2)
	    << "#nodes = " << (double)_noOfNodes << std::endl
	    << "#neighbours = " << _noOfNeighbours << std::endl
	    << "#edges = " << (double)_noOfEdges << std::endl
	    << "#nodes which fit in internal memory = "
	    << (double)_noOfNodesInInternalMemory << std::endl
	    << "--------------------" << std::endl << std::fixed;
    }

    virtual EdgeCount noOfEdges() const {return _noOfEdges;}
    virtual NodeCount noOfNodes() const {return _noOfNodes;}
    
 private:
    NodeCount _noOfNodes;
    NodeCount _noOfNeighbours;
    EdgeCount _noOfEdges;
};

class LoggingImportedGraph : public LoggingProblemInstance
{
 public:
    LoggingImportedGraph(NodeCount noOfNodes, EdgeCount noOfEdges,
			 NodeCount noOfNodesInInternalMemory,
			 std::string filename)
	: LoggingProblemInstance(noOfNodesInInternalMemory),
	  _noOfNodes( noOfNodes ), _noOfEdges( noOfEdges ), _filename( filename )
	{}
    
    virtual void printProblemInstance(std::ostream &out) const {
	out << std::endl << "IMPORTED GRAPH" << std::endl << std::endl
	    << "filename = '" << _filename << "'" << std::endl
	    << std::scientific << std::setprecision(2)
	    << "#nodes = " << (double)_noOfNodes << std::endl
	    << "#edges = " << (double)_noOfEdges << std::endl
	    << "#nodes which fit in internal memory = "
	    << (double)_noOfNodesInInternalMemory << std::endl
	    << "--------------------" << std::endl << std::fixed;
    }

    virtual EdgeCount noOfEdges() const {return _noOfEdges;}
    virtual NodeCount noOfNodes() const {return _noOfNodes;}
    
 private:
    NodeCount _noOfNodes;
    EdgeCount _noOfEdges;
    std::string _filename;
};



class LoggingEvent
{
 public:
    LoggingEvent( std::string name )
	: _name( name )
	{_timestamp = stxxl::stxxl_timestamp();}

    std::string name() const {return _name;}
    double timestamp() const {return _timestamp;}
    
 private:
    std::string _name;
    double _timestamp;
};

class LoggingEvents : public std::vector<LoggingEvent>
{
    friend std::ostream& operator<<( std::ostream& os, const LoggingEvents &events ) {
	if (events.size() > 0) {
	    os << std::endl << "EVENTS" << std::endl;
	    os.precision(5);
	    double firstTimestamp = events[0].timestamp();
	    double lastTimestamp = events[0].timestamp();
	    for (int i=0; i<events.size(); i++) {
		os << "total time: " << std::setw(9)
		   << (events[i].timestamp()-firstTimestamp) << " sec., section time: " << std::setw(9)
		   << (events[i].timestamp()-lastTimestamp) << " sec.  -  "
		   << events[i].name() << std::endl;
		lastTimestamp = events[i].timestamp();
	    }
	    os << "--------------------" << std::endl;
	}
	return os;
    }

 public:
    double totalTime() const {
	return ( back().timestamp() - (*this)[1].timestamp() );
    }
};

class LoggingResult
{
    friend std::ostream& operator<<( std::ostream& os, const LoggingResult &result ) {
	os << std::endl << "RESULT" << std::endl
	   << "Minimum Spanning Tree:" << std::endl
	   << "total weight = " << result._totalWeight << std::endl
	   << "#edges = " << result._noOfEdges << std::endl
	   << "--------------------" << std::endl;
	return os;
    }
    
 public:
    LoggingResult( EdgeCount noOfEdges, EdgeWeightBig totalWeight )
	: _noOfEdges( noOfEdges ), _totalWeight( totalWeight )
	{}

    EdgeCount noOfEdges() const {return _noOfEdges;}
    
 private:
    EdgeCount _noOfEdges;
    EdgeWeightBig _totalWeight;
};

class LoggingBucket
{
 public:
    LoggingBucket( Buckets<>::BucketID id, NodeCount noOfNodes, EdgeCount noOfEdges )
	: _id( id ), _noOfNodes( noOfNodes ), _noOfEdges( noOfEdges ),
	  _edgesProcessed( 0 ), _totalCapacity( 0 ), _minBlocksFree( -1 ),
	  _usedMemoryKB( 0 ), _usedMemoryPercent( 0 )
	{}

    void setBlocksUsed(int blocksUsed) {
	EdgeCount currentCapacity = (_noOfNodes + blocksUsed) * DS_INT_EDGES_PER_BLOCK;
	if (currentCapacity > _totalCapacity) _totalCapacity = currentCapacity;
    }

    void setBlocksFree(int blocksFree) {
	if ( (_minBlocksFree == -1) || (blocksFree < _minBlocksFree) )
	    _minBlocksFree = blocksFree;
    }

    void addEdgesProcessed(EdgeCount edgesProcessed) {
	_edgesProcessed += edgesProcessed;
    }
    
    Buckets<>::BucketID _id;
    NodeCount _noOfNodes;
    EdgeCount _noOfEdges;
    EdgeCount _edgesProcessed;
    EdgeCount _totalCapacity;
    int _minBlocksFree;
    int _usedMemoryKB;
    int _usedMemoryPercent;
};

class LoggingBuckets : public std::vector<LoggingBucket>
{
    friend std::ostream& operator<<( std::ostream& os, LoggingBuckets &buckets ) {
	if (buckets.size() > 0) {
	    buckets.readMemoryLog();
	    os << std::endl << "BUCKETS" << std::endl;
	    double maxFactor = 0;
	    EdgeCount maxTotalCapacity = 0;
	    for (Buckets<>::BucketID i=0; i<buckets.size(); i++) {
		
		double factor = buckets[i]._totalCapacity / (double)buckets[i]._noOfEdges;
		if (factor > maxFactor) maxFactor = factor;
		if (buckets[i]._totalCapacity > maxTotalCapacity)
		    maxTotalCapacity = buckets[i]._totalCapacity;
		
		os << std::setw(3) << buckets[i]._id
		   << std::scientific << std::setprecision(1)
		   << " #nodes=" << std::setw(5) << (double)buckets[i]._noOfNodes
		   << " #edges=" << std::setw(5) << (double)buckets[i]._noOfEdges
		   << " #edgesProcessed=" << std::setw(5) << (double)buckets[i]._edgesProcessed
		   << " capacity=" << std::setw(5) << (double)buckets[i]._totalCapacity
		   << " (factor " << std::fixed << factor << ")" << std::scientific
		   << " #free blocks=" << std::setw(5) << (double)buckets[i]._minBlocksFree
		   << " used memory=" << (buckets[i]._usedMemoryKB / 1024) << " MB"
		   << " (" << buckets[i]._usedMemoryPercent << "%)" << std::endl
		   << std::fixed;
	    }
	    os << std::endl << "max total capacity = " << maxTotalCapacity
	       << "  /  max factor = " << maxFactor << std::endl
	       << "--------------------" << std::endl;
	}
	return os;
    }

 public:
    void readMemoryLog() {
	std::ifstream inFile( "memoryLog.tmp" );
	if ( ! inFile ) {
	    std::cerr << "Wasn't able to open file 'memoryLog.tmp' in order to "
		      << "read the memory log." << std::endl;
	    return;
	}
	for (iterator iter=begin(); iter!=end(); iter++) {
	    double tmp;
	    inFile >> iter->_usedMemoryKB >> tmp;
	    iter->_usedMemoryPercent = tmp;
	}
	system( "rm memoryLog.tmp" );
    }
};

class LoggingDuplicates
{
    friend std::ostream& operator<<( std::ostream& os, const LoggingDuplicates &duplicates ) {
	os << std::endl << "DUPLICATES" << std::endl
	   << "#duplicates removed = " << duplicates.noOfDuplicates() << std::endl
	   << "--------------------" << std::endl;
	return os;
    }
    
 public:
    LoggingDuplicates()
	: _noOfDuplicates( 0 )
	{}

    void increase() {_noOfDuplicates++;}
    EdgeCount noOfDuplicates() const {return _noOfDuplicates;}
    
 private:
    EdgeCount _noOfDuplicates; 
};


class Logging
{
 public:
    Logging(std::string filename)
	: _filename( filename )
	{}
    
    ~Logging() {
	delete _problemInstance;
	delete _result;
    }

    void setProblemInstance(LoggingProblemInstance *problemInstance) {
	_problemInstance = problemInstance;
    }

    LoggingEvents &events() {return _events;}

    LoggingBuckets &buckets() {return _buckets;}

    void setResult(LoggingResult *result) {_result = result;}

    LoggingDuplicates &duplicates() {return _duplicates;}

    void printLog() {
	std::ofstream outFile( _filename.c_str() );
	if ( ! outFile ) {
	    std::cerr << "Wasn't able to open file '" << _filename
		      << "' in order to write the log." << std::endl;
	    return;
	}
	
	printConfig(outFile);
	printDotStxxl(outFile);
	_problemInstance->printProblemInstance(outFile);
	outFile << _events;
	DS_LOGGING2( outFile << _buckets );
	if (_result) outFile << *_result;
	DS_LOGGING2( printEdgesProcessed(outFile) );
	printTiming(outFile);
	DS_LOGGING2( DS_REMOVE_DUPLICATES( outFile << _duplicates ) );
    }
    
 private:
    std::string _filename;

    LoggingProblemInstance *_problemInstance;
    LoggingEvents _events;
    LoggingBuckets _buckets;
    LoggingResult *_result;
    LoggingDuplicates _duplicates;


    void printConfig(std::ostream &out)	const {
	out << "CONFIGURATION" << std::endl << std::endl
	    << "internal memory for sorting = "
	    << DS_INTERNAL_MEMORY_FOR_SORTING / (1024*1024) << " MB"
	    << std::endl << std::endl
	    
	    << "EdgeVector default block size = "
	    << DS_DEFAULT_BLOCK_SIZE / 1024 << " KB" << std::endl
	    << "EdgeVector default #pages = "
	    << DS_DEFAULT_NO_OF_PAGES << std::endl
	    << "EdgeVector default page size = "
	    << DS_DEFAULT_PAGE_SIZE << " block(s)" << std::endl << std::endl
	    
	    << "MST vector block size = "
	    << DS_MST_BLOCK_SIZE / 1024 << " KB" << std::endl
	    << "MST vector #pages = "
	    << DS_MST_NO_OF_PAGES << std::endl
	    << "MST vector page size = "
	    << DS_MST_PAGE_SIZE << " block(s)" << std::endl << std::endl
		   
	    << "Duplicates are"
	    DS_DONT_REMOVE_DUPLICATES(<< " NOT")
	    << " removed." << std::endl << std::endl

	    << "The input is"
	    DS_DONT_RANDOMIZE(<< " NOT")
	    << " randomized." << std::endl << std::endl

	    DS_USE_BUCKETS( << "buckets implementation" << std::endl << std::endl
			    << "external bucket block size = "
			    << DS_EXT_BLOCK_SIZE / 1024 << " KB" << std::endl
			    << "external bucket page size = "
			    << DS_EXT_PAGE_SIZE << " block(s)" << std::endl
			    << "external bucket prefetch pool size = "
			    << DS_EXT_PREFETCH_POOL << " block(s)" << std::endl
			    << "external bucket write pool size = "
			    << DS_EXT_WRITE_POOL << " block(s)" << std::endl << std::endl
	    
			    << "internal buckets: no of blocks = "
			    << DS_INT_NO_OF_BLOCKS << std::endl
			    << "internal buckets: edges per block = "
			    << DS_INT_EDGES_PER_BLOCK << std::endl
		)
	    
	    DS_USE_PQUEUE( << "priority queue implementation" << std::endl << std::endl 
			   << "priority queue internal memory = "
			   << DS_PQUEUE_INTERNAL_MEMORY / (1024*1024) << " MB" << std::endl
			   << "priority queue max size = "
			   << DS_PQUEUE_MAX_SIZE << std::endl
			   << "priority queue prefetch pool size = "
			   << DS_PQUEUE_PREFETCH_POOL << " block(s)" << std::endl
			   << "priority queue write pool size = "
			   << DS_PQUEUE_WRITE_POOL << " block(s)" << std::endl
		)
			   
	    << "--------------------" << std::endl;
    }


    void printDotStxxl(std::ostream &out) const {
	std::ifstream inFile( ".stxxl" );
	if ( ! inFile ) {
	    std::cerr << "Wasn't able to open file '.stxxl' in order to "
		      << "read the stxxl-disk-configuration." << std::endl;
	    return;
	}
	out << std::endl << "STXXL-DISK-CONFIGURATION (.stxxl)" << std::endl << std::endl;
	char ch;
	while ( inFile.get(ch) ) out.put(ch);
	out << std::endl << "--------------------" << std::endl;
    }

    void printEdgesProcessed(std::ostream &out) const {

	DS_USE_PQUEUE( return );
	
	if (_problemInstance->noOfNodes() <=
	    _problemInstance->noOfNodesInInternalMemory()) return;
	
	// compute the number of edges read or written during the reducing of nodes
	// and the total number of processed edges
	
	// 1) read edges from input
	double edgesIO = _problemInstance->noOfEdges();
	double edgesProcessed = 0;
	for (Buckets<>::BucketID i=0; i<_buckets.size(); i++) {
	    // 2) write edges to external buckets
	    edgesIO += _buckets[i]._noOfEdges;
	    // 3) read edges from external buckets (the first bucket isn't read during this step)
	    if (_buckets[i]._id != -1) edgesIO += _buckets[i]._noOfEdges;

	    // processed edges
	    edgesProcessed += _buckets[i]._edgesProcessed;
	}

	// compute the expected number of processed edges
	double expectedEdgesProcessed = 0;
	for (NodeCount u = _problemInstance->noOfNodesInInternalMemory();
	     u <= _problemInstance->noOfNodes();
	     u++) {
	    expectedEdgesProcessed += 2 * (double)_problemInstance->noOfEdges() / (double)(u+1);
	}
	
	out << std::endl << "PROCESSED EDGES" << std::endl << std::endl
	    << "reducing of nodes:" << std::endl
	    << "#edges read or written = "
	    << std::scientific << std::setprecision(3) << edgesIO
	    << " (" << (int)( edgesIO / 1024 * 20 / 1024 ) << " MB)" << std::endl
	    << "#edges processed = " << edgesProcessed << std::endl
	    << "expected #edges processed = " << expectedEdgesProcessed
	    << std::fixed << std::endl
	    << "--------------------" << std::endl;
    }

    void printTiming(std::ostream &out) const {
	double t = _events.totalTime();
	EdgeCount m = _problemInstance->noOfEdges();
	
	out << std::endl << "TIMING" << std::endl << std::endl
	    << "total time = " << std::scientific << std::setprecision(5)
	    << t << " sec." << std::endl
	    << "time / m = " << (t / m) << " sec." << std::endl
	    << "time / (m ln m) = " << (t / (m * log((double)m)))
	    << " sec." << std::fixed << std::endl
	    << "[m = #edges] " << std::endl
	    << "--------------------" << std::endl;
    }
};


#endif // LOGGING_H
