/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                parameters.h                      *
 *                                                  *
 *  July 13, 2003                                   *
 ****************************************************/


#ifndef PARAMETERS_H
#define PARAMETERS_H


/**
   Parses the command-line arguments and provides methods to get the arguments
   in a convenient way.
*/
class Parameters
{
 public:
    Parameters( int argc, char *argv[] )
	: _inputSpecified( false ),
	  _randomGraph( false ), _gridGraph( false ), _geometricGraph( false ),
	  _noOfNodesInInternalMemory( DS_EXT_FIRST_BUCKET_SIZE ),
	  _importInputFilename( "" ), _outputFilename( "" ), _exportInputFilename( "" ),
	  _reducedGraphFilename( "" ),
	  _exportInputFormatAdjacencyList( false ), _exportInputFormatCompressed( false ),
	  _quitAfterExporting( false ),
	  _abortAfterNodeReducing( false )
	
    {
	if (argc < 2) {
	    std::cerr << "Too few arguments !" << std::endl << std::endl;
	    usage(-1);
	}
	int index = 1;
	while (index < argc) {
	    if ( ! strcmp(argv[index], "-h") ) {
		usage(0);
	    }
	    
	    if ( ! strcmp(argv[index], "-r") ) {
		if ( _inputSpecified ) {
		    std::cerr << "Argument '-r' contradicts another argument !"
			      << std::endl << std::endl;
		    usage(-1);
		}
		if ( argc - (++index) < 2 ) {
		    std::cerr << "-r requires two additional arguments !"
			      << std::endl << std::endl;
		    usage(-1);
		}
		_randomGraphNoOfNodes = atof(argv[index++]);
		_noOfNodes = _randomGraphNoOfNodes;
		_randomGraphNoOfEdges = atof(argv[index++]);
		_randomGraph = true;
		_inputSpecified = true;
		continue;
	    }
	    
	    if ( ! strcmp(argv[index], "-g") ) {
		if ( _inputSpecified ) {
		    std::cerr << "Argument '-g' contradicts another argument !"
			      << std::endl << std::endl;
		    usage(-1);
		}
		if ( argc - (++index) < 2 ) {
		    std::cerr << "-g requires two additional arguments !"
			      << std::endl << std::endl;
		    usage(-1);
		}
		_gridGraphNoOfNodesX = atof(argv[index++]);
		_gridGraphNoOfNodesY = atof(argv[index++]);
		_noOfNodes = _gridGraphNoOfNodesX * _gridGraphNoOfNodesY;
		_gridGraph = true;
		_inputSpecified = true;
		continue;
	    }
	    
	    if ( ! strcmp(argv[index], "-geo") ) {
		if ( _inputSpecified ) {
		    std::cerr << "Argument '-geo' contradicts another argument !"
			      << std::endl << std::endl;
		    usage(-1);
		}
		if ( argc - (++index) < 2 ) {
		    std::cerr << "-geo requires two additional arguments !"
			      << std::endl << std::endl;
		    usage(-1);
		}
		_geometricGraphNoOfNodes = atof(argv[index++]);
		_geometricGraphNoOfNeighbours = atof(argv[index++]);
		_noOfNodes = _geometricGraphNoOfNodes;
		_geometricGraph = true;
		_inputSpecified = true;
		continue;
	    }

	    if ( ! strcmp(argv[index], "-i") ) {
		if ( _inputSpecified ) {
		    std::cerr << "Argument '-i' contradicts another argument !"
			      << std::endl << std::endl;
		    usage(-1);
		}
		if ( argc - (++index) < 1 ) {
		    std::cerr << "-i requires one additional argument !"
			      << std::endl << std::endl;
		    usage(-1);
		}
		_importInputFilename = argv[index++];
		_noOfNodes = noOfNodesInFile(_importInputFilename);
		_inputSpecified = true;
		continue;
	    }
	    
	    if ( ! strcmp(argv[index], "-a") ) {
		_abortAfterNodeReducing = true;
		if ( argc - (++index) >= 1 ) {
		    char *pchar = argv[index];
		    if (pchar[0] != '-') {
			_reducedGraphFilename = pchar;
			index++;
		    }
		}
		continue;
	    }
	    
	    if ( ! strcmp(argv[index], "-o") ) {
		if ( argc - (++index) < 1 ) {
		    std::cerr << "-o requires one additional argument !"
			      << std::endl << std::endl;
		    usage(-1);
		}
		_outputFilename = argv[index++];
		continue;
	    }

	    if ( ! strcmp(argv[index], "-m") ) {
		if ( argc - (++index) < 1 ) {
		    std::cerr << "-m requires one additional argument !"
			      << std::endl << std::endl;
		    usage(-1);
		}
		_noOfNodesInInternalMemory = atof(argv[index++]);
		continue;
	    }

	    if ( ! strcmp(argv[index], "-e") ) {
		if ( argc - (++index) < 1 ) {
		    std::cerr << "-e requires one additional argument !"
			      << std::endl << std::endl;
		    usage(-1);
		}
		_exportInputFilename = argv[index++];
		while (index < argc) {
		    if ( ! strcmp(argv[index], "-f") ) {
			if ( argc - (++index) < 1 ) {
			    std::cerr << "-f requires one additional argument !"
				      << std::endl << std::endl;
			    usage(-1);
			}
			if ( ! strcmp(argv[index], "a") )
			    _exportInputFormatAdjacencyList = true;
			if ( ! strcmp(argv[index++], "c") )
			    _exportInputFormatCompressed = true;
			continue;
		    }
		    if ( ! strcmp(argv[index], "-q") ) {
			_quitAfterExporting = true;
			index++;
			continue;
		    }
		    break;
		}
		continue;
	    }
	    
	    std::cerr << "Argument '" << argv[index] << "' unknown -> ignored." << std::endl;
	    index++;
	}
	
	if ( ! _inputSpecified ) {
	    std::cerr << "One of the arguments '-h', '-r', '-g', '-geo' "
		      << "or '-i' is required !"
		      << std::endl << std::endl;
	    usage(-1);
	}
    }

    NodeCount noOfNodes() const {return _noOfNodes;}
    NodeCount noOfNodesInInternalMemory() const {return _noOfNodesInInternalMemory;}
    
    bool randomGraph() const {return _randomGraph;}
    NodeCount randomGraphNoOfNodes() const {return _randomGraphNoOfNodes;}
    EdgeCount randomGraphNoOfEdges() const {return _randomGraphNoOfEdges;}

    bool gridGraph() const {return _gridGraph;}
    NodeCount gridGraphNoOfNodesX() const {return _gridGraphNoOfNodesX;}
    NodeCount gridGraphNoOfNodesY() const {return _gridGraphNoOfNodesY;}

    bool geometricGraph() const {return _geometricGraph;}
    NodeCount geometricGraphNoOfNodes() const {return _geometricGraphNoOfNodes;}
    NodeCount geometricGraphNoOfNeighbours() const {return _geometricGraphNoOfNeighbours;}

    std::string importInputFilename() const {return _importInputFilename;}
    std::string outputFilename() const {return _outputFilename;}
    std::string exportInputFilename() const {return _exportInputFilename;}
    std::string reducedGraphFilename() const {return _reducedGraphFilename;}
    bool exportInputFormatAdjacencyList() const {return _exportInputFormatAdjacencyList;}
    bool exportInputFormatCompressed() const {return _exportInputFormatCompressed;}
    bool quitAfterExporting() const {return _quitAfterExporting;}
    bool abortAfterNodeReducing() const {return _abortAfterNodeReducing;}
	
 private:
    bool _inputSpecified;
    NodeCount _noOfNodes;
    NodeCount _noOfNodesInInternalMemory;
    
    bool _randomGraph;
    NodeCount _randomGraphNoOfNodes;
    EdgeCount _randomGraphNoOfEdges;

    bool _gridGraph;
    NodeCount _gridGraphNoOfNodesX;
    NodeCount _gridGraphNoOfNodesY;

    bool _geometricGraph;
    NodeCount _geometricGraphNoOfNodes;
    NodeCount _geometricGraphNoOfNeighbours;

    std::string _importInputFilename;
    std::string _outputFilename;
    std::string _exportInputFilename;
    std::string _reducedGraphFilename;
    bool _exportInputFormatAdjacencyList;
    bool _exportInputFormatCompressed;
    bool _quitAfterExporting;
    bool _abortAfterNodeReducing;
    

    void usage(int exitCode) const {
	std::cout << "usage: mst -h|(-r #nodes #edges)|(-g #nodesX #nodesY)|"
		  << "(-geo #nodes #neighbours)|(-i filenameImportInput) "
		  << std::endl << "           [-m #nodesInIntMem] "
		  << "[-e filenameExportInput [-f a|c|e] [-q]]" << std::endl
		  << "           [-a [filenameReducedGraph]] [-o filenameMST]"
		  << std::endl << std::endl
		  << "\t -h   display this help and exit" << std::endl
		  << "\t -r   generate random graph with the given numbers of nodes and edges" << std::endl
		  << "\t -g   generate grid graph with '#nodesX' * '#nodesY' nodes"
		  << std::endl
		  << "\t -geo generate geometric graph with the given number of nodes; "
		  << std::endl
		  << "\t      each node is connected with the '#neighbours' nearest "
		  << "neighbours" << std::endl
		  << "\t -i   specify file in order to import the input graph"
		  << std::endl << std::endl
		  << "\t -m   specify the number of nodes which fit in internal memory [default: "
		  << DS_EXT_FIRST_BUCKET_SIZE << "]" << std::endl
		  << "\t -e   specify file in order to export the input graph" << std::endl
		  << "\t      -f  specify the format of the exported graph: a = adjacency list, c = compressed, "
		  << "e = list of edges [default]" << std::endl
		  << "\t      -q  quit after exporting the input graph" << std::endl
		  << "\t -a   abort after the 'node reducing'-phase; "
		  << "the reduced graph is saved if a filename is specified" << std::endl
		  << "\t -o   specify output file for the resulting mst"
		  << std::endl << std::endl;
	exit( exitCode );
    }
};

#endif // PARAMETERS_H
