/****************************************************
 *  Bachelor Project:                               *
 *  External Memory Minimum Spanning Trees          *
 *  by Dominik Schultes <mail@dominik-schultes.de>  *
 *                                                  *
 *                randomizer.h                      *
 *                                                  *
 *  May 22, 2003                                    *
 ****************************************************/


#ifndef RANDOMIZER_H
#define RANDOMIZER_H

/**
   A bijection that uses the linear congruential method.
*/
class RandomizerLinearCongruence
{
 public:
    /**
       The constructor.
       A bijection over {0,...p-1} is initialized, whereat p is the
       next prime number greater than or equal to a given n.
    */
    RandomizerLinearCongruence(NodeCount noOfNodes)
	{determineNextPrime(noOfNodes);}

    /** The bijection. */
    NodeID operator()(NodeID nodeID) const {
	typedef unsigned long long int bigNumber;
	
	bigNumber x = nodeID;
	const bigNumber a = 321321;
	const bigNumber c = 1;
	bigNumber m = _prime;

	// the linear congruential method
	return (a * x + c) % m;
    }
    
 private:
    /** The next prime number >= the given number of nodes. */
    NodeCount _prime;

    /** Determines the next prime number >= the given number of nodes. */
    void determineNextPrime(NodeCount noOfNodes) {
	NodeCount p;
	for (p = noOfNodes-1; ! isPrime(p); p++);
	_prime = p;
    }

    /** Returns true iff the given number is prime. */
    bool isPrime(NodeCount p) const {
	for (NodeCount i=2; i <= sqrt((double)p); i++) {
	    if (p % i == 0) return false;
	}
	return true;
    }
};


/**
   A bijection that uses Feistel permutations.
*/
class RandomizerFeistel
{
 public:
    /**
       The constructor.
       A bijection over {0,...,r^2-1} is initialized, whereat r is the
       next integer greater than or equal to the square root of a given n.       
    */
    RandomizerFeistel(NodeCount noOfNodes) {
	// compute the next integer >= the square root of the given number of nodes
	_sqRoot = (NodeID) sqrt((double)noOfNodes);
	if ((NodeCount)_sqRoot * _sqRoot < noOfNodes) _sqRoot++;
	// initialize the table of random numbers (used by the Feistel permutations)
	initRandomNumbers();
    }

    /** The bijection. */
    NodeID operator()(NodeID nodeID) const {
	// Split the given number into two parts
	int a = nodeID / _sqRoot;
	int b = nodeID % _sqRoot;

	// Perform the Feistel permutations
	for (int i=0; i<_noOfIterations; i++) {
	    int newA = b;
	    b = a + randomNumbers[i][b];
	    if (b >= _sqRoot) b -= _sqRoot; // mod _sqRoot
	    a = newA;
	}

	// Recombine a and b to obtain the result
	return a * _sqRoot + b;
    }
	
 private:
    /** The number of performed Feistel permutations. */
    static const int _noOfIterations = 2;
    
    /** The maximum size of _sqRoot. */
    static const int _maxSqRoot = 0x10000;

    /** The next integer >= the square root of the given number of nodes. */
    NodeID _sqRoot;

    /** The table of random numbers (used by the Feistel permutations). */
    int randomNumbers[_noOfIterations][_maxSqRoot];

    /** Initializes the table of random numbers. */
    void initRandomNumbers() {
	for (int i=0; i<_noOfIterations; i++)
	    for (int j=0; j<_maxSqRoot; j++) {
		// each random number is an integer in {0,...,_sqRoot-1}
		randomNumbers[i][j] = (int)(rand() / (double)(RAND_MAX+1.0) * _sqRoot);
	    }
    }
};


/**
   A bijection over {0,...,n-1} that can be used to randomize an Edge
   in order to obtain a (pseudo-)random permutation.
   Either RandomizerLinearCongruence or RandomizerFeistel can be used
   as underlying bijection.
*/
template <typename Bijection = RandomizerLinearCongruence>
class Randomizer
{
 public:
    /** The constructor. */
    Randomizer(NodeCount noOfNodes)
	: _noOfNodes( noOfNodes ), _bijection( noOfNodes )
	{}

    /**
       Randomizes an Edge. The source and the target vertices are randomized
       using the bijection. The original source and target vertices are
       saved, so a RelabeledEdge, which contains both the randomized and the
       original vertices, is returned.
    */
    RelabeledEdge randomize(const Edge &edge) const {
	return RelabeledEdge(randomize(edge.source()),randomize(edge.target()),
			     edge.weight(),edge.source(),edge.target());
    }
	
 private:
    /** The number of nodes that specifies the domain and co-domain of the bijection. */
    NodeCount _noOfNodes;

    /** The underlying bijection. */
    Bijection _bijection;

    /** Randomizes a node ID using the underlying bijection. */
    NodeID randomize(NodeID nodeID) const {
	NodeID x = nodeID;
	
	// The application of the underlying bijection is repeated
	// until the result is in the correct co-domain.
	do {
	    x = _bijection( x );
	} while (x >= _noOfNodes);
	
	return x;
    }
    
};

#endif // RANDOMIZER_H
